import asyncio
import enum
import iterm2.connection
import iterm2.notifications
import json
import typing

class VariableScopes(enum.Enum):
    """Describes the scope in which a variable can be evaluated."""
    SESSION = iterm2.api_pb2.VariableScope.Value("SESSION")  #: Session scope
    TAB = iterm2.api_pb2.VariableScope.Value("TAB")  #: Tab scope
    WINDOW = iterm2.api_pb2.VariableScope.Value("WINDOW")  #: Window scope
    APP = iterm2.api_pb2.VariableScope.Value("APP")  #: Whole-app scope

class VariableMonitor:
    """
    Watches for changes to a variable.

    `VariableMonitor` is a context manager that helps observe changes in iTerm2 Variables.

   :param connection: The connection to iTerm2.
   :param scope: The scope in which the variable should be evaluated.
   :param name: The variable name.
   :param identifier: A tab, window, or session identifier. Must correspond to the passed-in scope. If the scope is `APP` this should be None. If the scope is `SESSION` the identifier may be "all" or "active".

    .. seealso::
        * Example ":ref:`colorhost_example`"
        * Example ":ref:`theme_example`"

    Example:

      .. code-block:: python

          async with iterm2.VariableMonitor(
                  connection,
                  iterm2.VariableScopes.SESSION,
                  "jobName",
                  my_session.session_id) as mon:
              while True:
                  new_value = await mon.async_get()
                  DoSomething(new_value)
        """
    def __init__(
            self,
            connection: iterm2.connection.Connection,
            scope: VariableScopes,
            name: str,
            identifier: typing.Optional[str]):
        self.__connection = connection
        self.__scope = scope
        self.__name = name
        self.__identifier = identifier
        self.__queue = asyncio.Queue(loop=asyncio.get_event_loop())

    async def __aenter__(self):
        async def callback(_connection, message):
            """Called when a variable changes."""
            await self.__queue.put(message)

        self.__token = await iterm2.notifications.async_subscribe_to_variable_change_notification(
                self.__connection,
                callback,
                self.__scope.value,
                self.__name,
                self.__identifier)
        return self

    async def async_get(self) -> typing.Any:
        """Returns the new value of the variable."""
        result = await self.__queue.get()
        jsonNewValue = result.json_new_value
        return json.loads(jsonNewValue)

    async def __aexit__(self, exc_type, exc, _tb):
        await iterm2.notifications.async_unsubscribe(self.__connection, self.__token)




"""Provides access to color presets."""

import iterm2.connection
import iterm2.color
import iterm2.rpc
import typing

class ListPresetsException(Exception):
    """Something went wrong listing presets."""
    pass

class GetPresetException(Exception):
    """Something went wrong fetching a color preset."""
    pass

class ColorPreset:
    class Color(iterm2.color.Color):
        """Derives from :class:`~iterm2.Color`.

        Note this is `~iterm2.ColorPreset.Color`, despite how Sphinx chooses to format this documentation."""
        def __init__(self, r, g, b, a, color_space, key):
            super().__init__(r, g, b, a, color_space)
            self.__key = key

        @property
        def key(self) -> str:
            """Describes which property this color affects."""
            return self.__key

        def __repr__(self):
            return "({},{},{},{},{} {})".format(
                self.__key,
                round(self.red),
                round(self.green),
                round(self.blue),
                round(self.alpha),
                self.color_space)

    @staticmethod
    async def async_get_list(connection: iterm2.connection.Connection) -> typing.List[str]:
        """Fetches a list of color presets.

        :param connection: An :class:`~iterm2.Connection`.

        :returns: Names of the color presets.

        .. seealso::
            * Example ":ref:`current_preset_example`"
            * Example ":ref:`random_color_example`"
        """
        result = await iterm2.rpc.async_list_color_presets(connection)
        if result.color_preset_response.status == iterm2.api_pb2.ColorPresetResponse.Status.Value("OK"):
            return list(result.color_preset_response.list_presets.name)
        else:
            raise GetPresetException(iterm2.api_pb2.ColorPresetResponse.Status.Name(result.color_preset_response.status))

    @staticmethod
    async def async_get(connection: iterm2.connection.Connection, name: str) -> typing.Union[None, 'ColorPreset']:
        """Fetches a color preset with the given name.

        :param connection: The connection to iTerm2.
        :param name: The name of the preset to fetch.

        :returns: Either a new preset or `None`.

        .. seealso::
            * Example ":ref:`colorhost_example`"
            * Example ":ref:`current_preset_example`"
            * Example ":ref:`theme_example`"
            * Example ":ref:`darknight`"
        """
        result = await iterm2.rpc.async_get_color_preset(connection, name)
        if result.color_preset_response.status == iterm2.api_pb2.ColorPresetResponse.Status.Value("OK"):
            return ColorPreset(result.color_preset_response.get_preset.color_settings)
        else:
            raise ListPresetsException(iterm2.api_pb2.ColorPresetResponse.Status.Name(result.color_preset_response.status))

    def __init__(self, proto):
        """Do not call this directly. Use :meth:async_get instead."""
        self.__values = []
        for setting in proto:
            self.__values.append(ColorPreset.Color(
                setting.red * 255,
                setting.green * 255,
                setting.blue * 255,
                setting.alpha * 255,
                iterm2.color.ColorSpace(setting.color_space),
                setting.key))

    @property
    def values(self) -> typing.List['ColorPreset.Color']:
        """Returns a list of color settings.

        :returns: The colors belonging to the preset.
        """
        return self.__values




"""Provides interfaces for getting and changing preferences (excluding
per-profile preferences; see the profile submodule for that)"""
import enum
import iterm2.rpc
import json
import typing

class PreferenceKey(enum.Enum):
    """Keys identifying particular preference settings."""
    OPEN_PROFILES_WINDOW_AT_START = "OpenBookmark"  #: Open the profiles window at startup?  Mutually exclusive with OPEN_DEFAULT_ARRANGEMENT_AT_START and RESTORE_ONLY_HOTKEY_AT_START.  Takes a boolean.
    OPEN_DEFAULT_ARRANGEMENT_AT_START = "OpenArrangementAtStartup"  #: Open default arrangement at startup?  Mutually exclusive with OPEN_PROFILES_WINDOW_AT_START and RESTORE_ONLY_HOTKEY_AT_START.  Takes a boolean.
    RESTORE_ONLY_HOTKEY_AT_START = "OpenNoWindowsAtStartup"  #: Restore only hotkey window at startup?  Mutually exclusive with OPEN_PROFILES_WINDOW_AT_START and OPEN_DEFAULT_ARRANGEMENT_AT_START.  Takes a boolean.
    QUIT_WHEN_ALL_WINDOWS_CLOSED = "QuitWhenAllWindowsClosed"  #: Quit automatically when all terminal windows are closed?  Takes a boolean.
    ONLY_WHEN_MORE_TABS = "OnlyWhenMoreTabs"  #: Confirm close window when there are multiple tabs?  Takes a boolean.
    PROMPT_ON_QUIT = "PromptOnQuit"  #: Prompt before quitting?  Takes a boolean.
    INSTANT_REPLAY_MEMORY_MB = "IRMemory"  #: Memory (in megabytes) to use per session for instant replay.  Takes a floating point value.
    SAVE_PASTE_HISTORY = "SavePasteHistory"  #: Should paste and command history be saved to disk?  Takes a boolean.
    ENABLE_BONJOUR_DISCOVERY = "EnableRendezvous"  #: Discover hosts with bonjour?  Takes a boolean.
    SOFTWARE_UPDATE_ENABLE_AUTOMATIC_CHECKS = "SUEnableAutomaticChecks"  #: Automatically check for new versions of iTerm2?  Takes a boolean.
    SOFTWARE_UPDATE_ENABLE_TEST_RELEASES = "CheckTestRelease"  #: Check for beta versions for auto update?  Takes a boolean.
    LOAD_PREFS_FROM_CUSTOM_FOLDER = "LoadPrefsFromCustomFolder"  #: Load prefs from a custom folder?  Takes a boolean.
    CUSTOM_FOLDER_TO_LOAD_PREFS_FROM = "PrefsCustomFolder"  #: If LOAD_PREFS_FROM_CUSTOM_FOLDER, gives the folder or URL to load prefs from.  Takes a string.
    COPY_TO_PASTEBOARD_ON_SELECTION = "CopySelection"  #: Copy to pasteboard on selection?  Takes a boolean.
    INCLUDE_TRAILING_NEWLINE_WHEN_COPYING = "CopyLastNewline"  #: Include trailing newline when copying to pasteboard?  Takes a boolean.
    APPS_MAY_ACCESS_PASTEBOARD = "AllowClipboardAccess"  #: Allow terminal apps to access the pasteboard?  Takes a boolean.
    WORD_CHARACTERS = "WordCharacters"  #: Characters considered part of a word for selection.  Takes a string.
    ENABLE_SMART_WINDOW_PLACEMENT = "SmartPlacement"  #: Enable smart window placement?  Takes a boolean.
    ADJUST_WINDOW_FOR_FONT_SIZE_CHANGE = "AdjustWindowForFontSizeChange"  #: Change window size when font size changes?  Takes a boolean.
    MAX_VERTICALLY = "MaxVertically"  #: When maximizing a window, grow it only vertically?  Takes a boolean.
    NATIVE_FULL_SCREEN_WINDOWS = "UseLionStyleFullscreen"  #: Use native full screen window?  Takes a boolean.
    OPEN_TMUX_WINDOWS_IN = "OpenTmuxWindowsIn"  #: Specifies how to open tmux windows.  Takes an integer. 0=native windows, 1=new window, 2=tabs in existing window.
    TMUX_DASHBOARD_LIMIT = "TmuxDashboardLimit"  #: Open tmux dashboard if there are more than this many windows.  Takes an integer.
    AUTO_HIDE_TMUX_CLIENT_SESSION = "AutoHideTmuxClientSession"  #: Automatically bury the tmux client session?  Takes a boolean.
    USE_TMUX_PROFILE = "TmuxUsesDedicatedProfile"  #: Use dedicated tmux profile. Takes a boolean.
    USE_METAL = "UseMetal"  #: Use the GPU renderer?  Takes a boolean.
    DISABLE_METAL_WHEN_UNPLUGGED = "disableMetalWhenUnplugged"  #: Disable the GPU renderer when not connected to power?  Takes a boolean.
    PREFER_INTEGRATED_GPU = "preferIntegratedGPU"  #: Prefer the integrated GPU over discrete, if available?  Takes a boolean.
    METAL_MAXIMIZE_THROUGHPUT = "metalMaximizeThroughput"  #: Maximize throughput for GPU renderer, vs framerate?  Takes a boolean.
    THEME = "TabStyleWithAutomaticOption"  #: Theme.  Takes an integer.  0 = Light, 1 = Dark, 2 = Light high contrast, 3 = Dark high contrast, 4 = Automatic (10.14+), 5 = Minimal (10.14+).
    TAP_BAR_POSTIION = "TabViewType"  #: Where the tab bar should be placed.  Takes an integer.  0=Top, 1=Bottom, 2=Left.
    HIDE_TAB_BAR_WHEN_ONLY_ONE_TAB = "HideTab"  #: Hide tab bar when there is only one tab?  Takes a boolean.
    HIDE_TAB_NUMBER = "HideTabNumber"  #: Hide the tab number?  Takes a boolean.
    HIDE_TAB_CLOSE_BUTTON = "HideTabCloseButton"  #: Hide the tab close button?  Takes a boolean.
    HIDE_TAB_ACTIVITY_INDICATOR = "HideActivityIndicator"  #: Hide the tab activity indicator?  Takes a boolean.
    SHOW_TAB_NEW_OUTPUT_INDICATOR = "ShowNewOutputIndicator"  #: Show a "new output" indicator in tabs?  Takes a boolean.
    SHOW_PANE_TITLES = "ShowPaneTitles"  #: Show a per-pane title bar?  Takes a boolean.
    STRETCH_TABS_TO_FILL_BAR = "StretchTabsToFillBar"  #: Stretch tabs horizontally to fill the tab bar?  Takes a boolean.
    HIDE_MENU_BAR_IN_FULLSCREEN = "HideMenuBarInFullscreen"  #: Hide menu bar when in full screen?  Takes a boolean.
    HIDE_FROM_DOCK_AND_APP_SWITCHER = "HideFromDockAndAppSwitcher"  #: Exclude iTerm2 from dock and app switcher?  Takes a boolean.
    FLASH_TAB_BAR_IN_FULLSCREEN = "FlashTabBarInFullscreen"  #: Flash tab bar in full screen?  Takes a boolean.
    WINDOW_NUMBER = "WindowNumber"  #: Show window number in title bar?  Takes a boolean.
    DIM_ONLY_TEXT = "DimOnlyText"  #: Dim only text when indicating inactive sessions?  Takes a boolean.
    SPLIT_PANE_DIMMING_AMOUNT = "SplitPaneDimmingAmount"  #: How much to dim inactive split panes or inactive windows.  Takes a floating point value in [0,1]
    DIM_INACTIVE_SPLIT_PANES = "DimInactiveSplitPanes"  #: Dim inactive split panes?  Takes a boolean.
    DRAW_WINDOW_BORDER = "UseBorder"  #: Show a border around windows?  Takes a boolean.
    HIDE_SCROLLBAR = "HideScrollbar"  #: Hide scroll bars?  Takes a boolean.
    DISABLE_FULLSCREEN_TRANSPARENCY = "DisableFullscreenTransparency"  #: Disable transparency for full screen windows?  Takes a boolean.
    ENABLE_DIVISION_VIEW = "EnableDivisionView"  #: Draw a line under the tab bar?  Takes a boolean.
    ENABLE_PROXY_ICON = "EnableProxyIcon"  #: Show proxy icon in title bar?  Takes a boolean.
    DIM_BACKGROUND_WINDOWS = "DimBackgroundWindows"  #: Dim inactive windows?  Takes a boolean.
    CONTROL_REMAPPING = "Control"  #: Remap control key.  Takes an integer.  1=Control, 2=Left option, 3=Right option, 7=Left Cmd, 8=Right Cmd.
    LEFT_OPTION_REMAPPING = "LeftOption"  #: Remap left option key.  Takes an integer.  1=Control, 2=Left option, 3=Right option, 7=Left Cmd, 8=Right Cmd.
    RIGHT_OPTION_REMAPPING = "RightOption"  #: Remap right option key.  Takes an integer.  1=Control, 2=Left option, 3=Right option, 7=Left Cmd, 8=Right Cmd.
    LEFT_COMMAND_REMAPPING = "LeftCommand"  #: Remap left cmd key.  Takes an integer.  1=Control, 2=Left option, 3=Right option, 7=Left Cmd, 8=Right Cmd.
    RIGHT_COMMAND_REMAPPING = "RightCommand"  #: Remap right cmd key.  Takes an integer.  1=Control, 2=Left option, 3=Right option, 7=Left Cmd, 8=Right Cmd.
    SWITCH_PANE_MODIFIER = "SwitchPaneModifier"  #: Modifiers to switch split pane with number.  Takes an integer.  3=Cmd, 6=Cmd+Option, 5=Option, 9=Disable switching.
    SWITCH_TAB_MODIFIER = "SwitchTabModifier"  #: Modifiers to switch tab by number.  Takes an integer.  3=Cmd, 6=Cmd+Option, 5=Option, 9=Disable switching.
    SWITCH_WINDOW_MODIFIER = "SwitchWindowModifier"  #: Modifiers to switch window with number.  Takes an integer.  3=Cmd, 6=Cmd+Option, 5=Option, 9=Disable switching.
    ENABLE_SEMANTIC_HISTORY = "CommandSelection"  #: Enable semantic history?  Takes a boolean.
    PASS_ON_CONTROL_CLICK = "PassOnControlClick"  #: Pass control-click to mouse reporting?  Takes a boolean.
    OPTION_CLICK_MOVES_CURSOR = "OptionClickMovesCursor"  #: Opt-click moves cursor?  Takes a boolean.
    THREE_FINGER_EMULATES = "ThreeFingerEmulates"  #: Three-finger tap emulates middle click?  Takes a boolean.
    FOCUS_FOLLOWS_MOUSE = "FocusFollowsMouse"  #: Focus follows mouse?  Takes a boolean.
    TRIPLE_CLICK_SELECTS_FULL_WRAPPED_LINES = "TripleClickSelectsFullWrappedLines"  #: Triple click selects full wrapped lines?  Takes a boolean.
    DOUBLE_CLICK_PERFORMS_SMART_SELECTION = "DoubleClickPerformsSmartSelection"  #: Double click performs smart selection?  Takes a boolean.
    ITERM_VERSION = "iTerm Version"  #: Last-used iTerm2 version. Do not set this.  Takes a string.
    AUTO_COMMAND_HISTORY = "AutoCommandHistory"  #: Enable autocomplete with command history?  Takes a boolean.
    PASTE_SPECIAL_CHUNK_SIZE = "PasteSpecialChunkSize"  #: Default paste chunk size.  Takes a positive integer.
    PASTE_SPECIAL_CHUNK_DELAY = "PasteSpecialChunkDelay"  #: Default delay between paste chunks.  Takes a floating point number.
    NUMBER_OF_SPACES_PER_TAB = "NumberOfSpacesPerTab"  #: Default number of spaces per tab when converting tabs when pasting.  Takes a positive integer.
    TAB_TRANSFORM = "TabTransform"  #: How to ransform tabs to space by default when pasting Takes an integer.  0=No transformation, 1=Convert to spaces, 2=Escape with C-V
    ESCAPE_SHELL_CHARS_WITH_BACKSLASH = "EscapeShellCharsWithBackslash"  #: Escape shell characters with backslash when using advanced paste?  Takes a boolean.
    CONVERT_UNICODE_PUNCTUATION = "ConvertUnicodePunctuation"  #: Convert unicode punctuation to ascii when using advanced paste?  Takes a boolean.
    CONVERT_DOS_NEWLINES = "ConvertDosNewlines"  #: Convert DOS newlines to Unix when using advanced paste?  Takes a boolean.
    REMOVE_CONTROL_CODES = "RemoveControlCodes"  #: Remove control codes when using advanced paste?  Takes a boolean.
    BRACKETED_PASTE_MODE = "BracketedPasteMode"  #: Allow bracketed paste mode when using advanced paste?  Takes a boolean.
    PASTE_SPECIAL_USE_REGEX_SUBSTITUTION = "PasteSpecialUseRegexSubstitution"  #: Enable regex substitution when using advanced paste?  Takes a boolean.
    PASTE_SPECIAL_REGEX = "PasteSpecialRegex"  #: Regex pattern to use for substitution when using advanced paste?  Takes a string.
    PASTE_SPECIAL_SUBSTITUTION = "PasteSpecialSubstitution"  #: Value to use for regex substitution when using advanced paste?  Takes a string.
    LEFT_TAB_BAR_WIDTH = "LeftTabBarWidth"  #: Width of left-side tab bar.  Takes a floating point value.
    PASTE_TAB_TO_STRING_TAB_STOP_SIZE = "PasteTabToStringTabStopSize"  #: When converting tabs to spaces, how many spaces to use?  Takes a nonnegative integer.
    SHOW_FULL_SCREEN_TAB_BAR = "ShowFullScreenTabBar"  #: Show tab bar in full screen?  Takes a boolean.
    DEFAULT_TOOLBELT_WIDTH = "Default Toolbelt Width"  #: Width of toolbelt by default.  Takes a nonnegative integer.
    SIZE_CHANGES_AFFECT_PROFILE = "Size Changes Affect Profile"  #: Does changing text size with cmd-+ and cmd-- affect only the session or also its profile?
    STATUS_BAR_POSITION = "StatusBarPosition"  #: Where does the status bar go? Takes an integer. 0=top, 1=bottom.
    PRESERVE_WINDOW_SIZE_WHEN_TAB_BAR_VISIBILITY_CHANGES = "PreserveWindowSizeWhenTabBarVisibilityChanges"  #: Keep window size the same when tabbar shows/hides? Takes a boolean.
    PER_PANE_BACKGROUND_IMAGE = "PerPaneBackgroundImage"  #: Should each split pane have a separate bg image, or one for the whole window? Takes a boolean.
    PER_PANE_STATUS_BAR = "SeparateStatusBarsPerPane"  #: Should each split pane have a separate status bar, or just one for the whole window? Takes a boolean.
    EMULATE_US_KEYBOARD = "UseVirtualKeyCodesForDetectingDigits"  #: Emulate US keyboard for the purposes of switching tabs/panes/windows by keyboard? Takes a boolean.
    TEXT_SIZE_CHANGES_AFFECT_PROFILE = "Size Changes Affect Profile"  #: Does increasing/decreasing text size update the backing profile? Takes a boolean.



async def async_get_preference(connection, key: PreferenceKey) -> typing.Union[None, typing.Any]:
    """
    Gets a preference by key.

    :param key: The preference key, from the `PreferenceKey` enum.
    :returns: An object with the preferences value, or `None` if unset and no default exists.
    """
    proto = await iterm2.rpc.async_get_preference(connection, key.value)
    j = proto.preferences_response.results[0].get_preference_result.json_value
    return json.loads(j)




"""Provides classes for representing, querying, and modifying iTerm2 profiles."""
import asyncio
import iterm2.color
import iterm2.colorpresets
import enum
import iterm2.rpc
import json
import typing

class BackgroundImageMode(enum.Enum):
    """Describes how the background image should be accommodated to fit the window."""
    STRETCH = 0  #: Stretch to fit
    TILE = 1  #: Full size, undistorted, and tessellated if needed.
    ASPECT_FILL = 2  #: Scale to fill the space, cropping if needed. Does not distort.
    ASPECT_FIT = 3  #: Scale to fit the space, adding letterboxes or pillarboxes if needed. Does not distort.

class BadGUIDException(Exception):
    """Raised when a profile does not have a GUID or the GUID is unknown."""

class CursorType(enum.Enum):
    """Describes the type of the cursor."""
    CURSOR_TYPE_UNDERLINE = 0  #: Underline cursor
    CURSOR_TYPE_VERTICAL = 1  #: Vertical bar cursor
    CURSOR_TYPE_BOX = 2  #: Box cursor

class ThinStrokes(enum.Enum):
    """When thin strokes should be used."""
    THIN_STROKES_SETTING_NEVER = 0  #: NEver
    THIN_STROKES_SETTING_RETINA_DARK_BACKGROUNDS_ONLY = 1  #: When the background is dark and the display is a retina display.
    THIN_STROKES_SETTING_DARK_BACKGROUNDS_ONLY = 2  #: When the background is dark.
    THIN_STROKES_SETTING_ALWAYS = 3  #: Always.
    THIN_STROKES_SETTING_RETINA_ONLY = 4  #: When the display is a retina display.

class UnicodeNormalization(enum.Enum):
    """How to perform Unicode normalization."""
    UNICODE_NORMALIZATION_NONE = 0  #: Do not modify input
    UNICODE_NORMALIZATION_NFC = 1  #: Normalization form C
    UNICODE_NORMALIZATION_NFD = 2  #: Normalization form D
    UNICODE_NORMALIZATION_HFSPLUS = 3  #: Apple's HFS+ normalization form

class CharacterEncoding(enum.Enum):
    """String encodings."""
    CHARACTER_ENCODING_UTF_8 = 4

class OptionKeySends(enum.Enum):
    """How should the option key behave?"""
    OPTION_KEY_NORMAL = 0  #: Standard behavior
    OPTION_KEY_META = 1  #: Acts like Meta. Not recommended.
    OPTION_KEY_ESC = 2  #: Adds ESC prefix.

class InitialWorkingDirectory(enum.Enum):
    """How should the initial working directory of a session be set?"""
    INITIAL_WORKING_DIRECTORY_CUSTOM = "Yes"  #: Custom directory, specified elsewhere
    INITIAL_WORKING_DIRECTORY_HOME = "No"  #: Use default of home directory
    INITIAL_WORKING_DIRECTORY_RECYCLE = "Recycle"  #: Reuse the "current" directory, or home if there is no current.
    INITIAL_WORKING_DIRECTORY_ADVANCED = "Advanced"  #: Use advanced settings, which specify more granular behavior depending on whether the new session is a new window, tab, or split pane.

class IconMode(enum.Enum):
    """How should session icons be selected?"""
    NONE = 0
    AUTOMATIC = 1
    CUSTOM = 2

class TitleComponents(enum.Enum):
    """Which title components should be present?"""
    SESSION_NAME = (1 << 0)
    JOB = (1 << 1)
    WORKING_DIRECTORy = (1 << 2)
    TTY = (1 << 3)
    CUSTOM = (1 << 4)  #: Mutually exclusive with all other options.
    PROFILE_NAME = (1 << 5)
    PROFILE_AND_SESSION_NAME = (1 << 6)
    USER = (1 << 7)
    HOST = (1 << 8)

class LocalWriteOnlyProfile:
    """A profile that can be modified but not read and does not send changes on each write.

    You can safely create this with `LocalWriteOnlyProfile()`. Use
    :meth:`~iterm2.Session.async_set_profile_properties` to update a session
    without modifying the underlying profile.

    .. seealso:: Example ":ref:`copycolor`"
    """
    def __init__(self):
      self.__values = {}

    @property
    def values(self):
        return self.__values

    def _simple_set(self, key, value):
        """value is a json type"""
        if key is None:
            self.__values[key] = None
        else:
            self.__values[key] = json.dumps(value)

    def _color_set(self, key, value):
        if value is None:
            self.__values[key] = "null"
        else:
            self.__values[key] = json.dumps(value.get_dict())

    def _guids_for_set(self):
        if self.session_id is None:
            assert self.__guid is not None
            return [self.__guid]
        else:
            return self.session_id

    def set_foreground_color(self, value: 'iterm2.color.Color'):
        """Sets the foreground color.

        :param value: A :class:`Color`"""
        return self._color_set("Foreground Color", value)

    def set_background_color(self, value: 'iterm2.color.Color'):
        """Sets the background color.

        :param value: A :class:`Color`"""
        return self._color_set("Background Color", value)

    def set_bold_color(self, value: 'iterm2.color.Color'):
        """Sets the bold text color.

        :param value: A :class:`Color`"""
        return self._color_set("Bold Color", value)

    def set_link_color(self, value: 'iterm2.color.Color'):
        """Sets the link color.

        :param value: A :class:`Color`"""
        return self._color_set("Link Color", value)

    def set_selection_color(self, value: 'iterm2.color.Color'):
        """Sets the selection background color.

        :param value: A :class:`Color`"""
        return self._color_set("Selection Color", value)

    def set_selected_text_color(self, value: 'iterm2.color.Color'):
        """Sets the selection text color.

        :param value: A :class:`Color`"""
        return self._color_set("Selected Text Color", value)

    def set_cursor_color(self, valu: 'iterm2.color.Color'):
        """Sets the cursor color.

        :param value: A :class:`Color`"""
        return self._color_set("Cursor Color", value)

    def set_cursor_text_color(self, value: 'iterm2.color.Color'):
        """Sets the cursor text color.

        :param value: A :class:`Color`"""
        return self._color_set("Cursor Text Color", value)

    def set_ansi_0_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 0 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 0 Color", value)

    def set_ansi_1_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 1 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 1 Color", value)

    def set_ansi_2_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 2 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 2 Color", value)

    def set_ansi_3_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 3 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 3 Color", value)

    def set_ansi_4_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 4 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 4 Color", value)

    def set_ansi_5_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 5 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 5 Color", value)

    def set_ansi_6_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 6 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 6 Color", value)

    def set_ansi_7_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 7 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 7 Color", value)

    def set_ansi_8_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 8 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 8 Color", value)

    def set_ansi_9_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 9 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 9 Color", value)

    def set_ansi_10_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 10 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 10 Color", value)

    def set_ansi_11_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 11 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 11 Color", value)

    def set_ansi_12_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 12 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 12 Color", value)

    def set_ansi_13_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 13 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 13 Color", value)

    def set_ansi_14_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 14 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 14 Color", value)

    def set_ansi_15_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 15 color.

        :param value: A :class:`Color`"""
        return self._color_set("Ansi 15 Color", value)

    def set_smart_cursor_color(self, value: 'iterm2.color.Color'):
        """Sets the smart cursor color.

        :param value: A :class:`Color`"""
        return self._simple_set("Smart Cursor Color", value)

    def set_tab_color(self, value: 'iterm2.color.Color'):
        """Sets the tab color.

        :param value: A :class:`Color`"""
        return self._color_set("Tab Color", value)

    def set_underline_color(self, value: 'iterm2.color.Color'):
        """Sets the underline color.

        :param value: A :class:`Color` or None"""
        return self._color_set("Underline Color", value)

    def set_cursor_guide_color(self, value: 'iterm2.color.Color'):
        """Sets the cursor guide color. The alpha value is respected.

        :param value: A :class:`Color`"""
        return self._color_set("Cursor Guide Color", value)

    def set_badge_color(self, value: 'iterm2.color.Color'):
        """Sets the badge color. The alpha value is respected.

        :param value: A :class:`Color`"""
        return self._color_set("Badge Color", value)

    def set_name(self, value: str):
        """Sets the name.

        :param value: A string"""
        return self._simple_set("Name", value)

    def set_badge_text(self, value: str):
        """Sets the badge text.

        :param value: A :class:`Color`"""
        return self._simple_set("Badge Text", value)

    def set_answerback_string(self, value: str):
        """Sets the answerback string.

        :param value: A string"""
        return self._simple_set("Answerback String", value)

    def set_use_cursor_guide(self, value: bool):
        """Sets whether the cursor guide should be used.

        :param value: A boolean"""
        return self._simple_set("Use Cursor Guide", value)

    def set_use_tab_color(self, value: bool):
        """Sets whether the tab color should be used.

        :param value: A boolean"""
        return self._simple_set("Use Tab Color", value)

    def set_use_underline_color(self, value: 'iterm2.color.Color'):
        """Sets the underline color.

        :param value: A :class:`Color`"""
        return self._simple_set("Use Underline Color", value)

    def set_minimum_contrast(self, value: float):
        """Sets the minimum contrast.

        :param value: A float in 0 to 1"""
        return self._simple_set("Minimum Contrast", value)

    def set_cursor_boost(self, value: float):
        """Sets the cursor boost level.

        :param value: A float in 0 to 1"""
        return self._simple_set("Cursor Boost", value)

    def set_blinking_cursor(self, value: bool):
        """Sets whether the cursor blinks.

        :param value: A bool"""
        return self._simple_set("Blinking Cursor", value)

    def set_use_bold_font(self, value: bool):
        """Sets whether to use the bold variant of the font for bold text.

        :param value: A bool"""
        return self._simple_set("Use Bold Font", value)

    def set_ascii_ligatures(self, value: bool):
        """Sets whether ligatures should be used for ASCII text.

        :param value: A bool"""
        return self._simple_set("ASCII Ligatures", value)

    def set_non_ascii_ligatures(self, value: bool):
        """Sets whether ligatures should be used for non-ASCII text.

        :param value: A bool"""
        return self._simple_set("Non-ASCII Ligatures", value)

    def set_use_bright_bold(self, value: bool):
        """Sets whether bright colors should be used for bold text.

        :param value: A bool"""
        return self._simple_set("Use Bright Bold", value)

    def set_blink_allowed(self, value: bool):
        """Sets whether blinking text is allowed.

        :param value: A bool"""
        return self._simple_set("Blink Allowed", value)

    def set_use_italic_font(self, value: bool):
        """Sets whether italic text is allowed.

        :param value: A bool"""
        return self._simple_set("Use Italic Font", value)

    def set_ambiguous_double_width(self, value: bool):
        """Sets whether ambiguous-width text should be treated as double-width.

        :param value: A bool"""
        return self._simple_set("Ambiguous Double Width", value)

    def set_horizontal_spacing(self, value: float):
        """Sets the fraction of horizontal spacing.

        :param value: A float at least 0"""
        return self._simple_set("Horizontal Spacing", value)

    def set_vertical_spacing(self, value: float):
        """Sets the fraction of vertical spacing.

        :param value: A float at least 0"""
        return self._simple_set("Vertical Spacing", value)

    def set_use_non_ascii_font(self, value: bool):
        """Sets whether to use a different font for non-ASCII text.

        :param value: A bool"""
        return self._simple_set("Use Non-ASCII Font", value)

    def set_transparency(self, value: float):
        """Sets the level of transparency.

        :param value: A float between 0 and 1"""
        return self._simple_set("Transparency", value)

    def set_blur(self, value: bool):
        """Sets whether background blur should be enabled.

        :param value: A bool"""
        return self._simple_set("Blur", value)

    def set_blur_radius(self, value: float):
        """Sets the blur radius (how blurry). Requires blur to be enabled.

        :param value: A float between 0 and 30"""
        return self._simple_set("Blur Radius", value)

    def set_background_image_mode(self, value: BackgroundImageMode):
        """Sets how the background image is drawn.

        :param value: A `BackgroundImageMode`"""
        return self._simple_set("Background Image Mode", value)

    def set_blend(self, value: float):
        """Sets how much the default background color gets blended with the background image.

        :param value: A float in 0 to 1"""
        return self._simple_set("Blend", value)

    def set_sync_title(self, value: bool):
        """Sets whether the profile name stays in the tab title, even if changed by an escape
        sequence.

        :param value: A bool"""
        return self._simple_set("Sync Title", value)

    def set_disable_window_resizing(self, value: bool):
        """Sets whether the terminal can resize the window with an escape sequence.

        :param value: A bool"""
        return self._simple_set("Disable Window Resizing", value)

    def set_only_the_default_bg_color_uses_transparency(self, value: bool):
        """Sets whether window transparency shows through non-default background colors.

        :param value: A bool"""
        return self._simple_set("Only The Default BG Color Uses Transparency", value)

    def set_ascii_anti_aliased(self, value: bool):
        """Sets whether ASCII text is anti-aliased.

        :param value: A bool"""
        return self._simple_set("ASCII Anti Aliased", value)

    def set_non_ascii_anti_aliased(self, value: bool):
        """Sets whether non-ASCII text is anti-aliased.

        :param value: A bool"""
        return self._simple_set("Non-ASCII Anti Aliased", value)

    def set_scrollback_lines(self, value: int):
        """Sets the number of scrollback lines.

        :param value: An int at least 0"""
        return self._simple_set("Scrollback Lines", value)

    def set_unlimited_scrollback(self, value: bool):
        """Sets whether the scrollback buffer's length is unlimited.

        :param value: A bool"""
        return self._simple_set("Unlimited Scrollback", value)

    def set_scrollback_with_status_bar(self, value: bool):
        """Sets whether text gets appended to scrollback when there is an app status bar

        :param value: A bool"""
        return self._simple_set("Scrollback With Status Bar", value)

    def set_scrollback_in_alternate_screen(self, value: bool):
        """Sets whether text gets appended to scrollback in alternate screen mode

        :param value: A bool"""
        return self._simple_set("Scrollback in Alternate Screen", value)

    def set_mouse_reporting(self, value: bool):
        """Sets whether mouse reporting is allowed

        :param value: A bool"""
        return self._simple_set("Mouse Reporting", value)

    def set_mouse_reporting_allow_mouse_wheel(self, value: bool):
        """Sets whether mouse reporting reports the mouse wheel's movements.

        :param value: A bool"""
        return self._simple_set("Mouse Reporting allow mouse wheel", value)

    def set_allow_title_reporting(self, value: bool):
        """Sets whether the session title can be reported

        :param value: A bool"""
        return self._simple_set("Allow Title Reporting", value)

    def set_allow_title_setting(self, value: bool):
        """Sets whether the session title can be changed by escape sequence

        :param value: A bool"""
        return self._simple_set("Allow Title Setting", value)

    def set_disable_printing(self, value: bool):
        """Sets whether printing by escape sequence is disabled.

        :param value: A bool"""
        return self._simple_set("Disable Printing", value)

    def set_disable_smcup_rmcup(self, value: bool):
        """Sets whether alternate screen mode is disabled

        :param value: A bool"""
        return self._simple_set("Disable Smcup Rmcup", value)

    def set_silence_bell(self, value: bool):
        """Sets whether the bell makes noise.

        :param value: A bool"""
        return self._simple_set("Silence Bell", value)

    def set_bm_growl(self, value: bool):
        """Sets whether notifications should be shown.

        :param value: A bool"""
        return self._simple_set("BM Growl", value)

    def set_send_bell_alert(self, value: bool):
        """Sets whether notifications should be shown for the bell ringing

        :param value: A bool"""
        return self._simple_set("Send Bell Alert", value)

    def set_send_idle_alert(self, value: bool):
        """Sets whether notifications should be shown for becoming idle

        :param value: A bool"""
        return self._simple_set("Send Idle Alert", value)

    def set_send_new_output_alert(self, value: bool):
        """Sets whether notifications should be shown for new output

        :param value: A bool"""
        return self._simple_set("Send New Output Alert", value)

    def set_send_session_ended_alert(self, value: bool):
        """Sets whether notifications should be shown for a session ending

        :param value: A bool"""
        return self._simple_set("Send Session Ended Alert", value)

    def set_send_terminal_generated_alerts(self, value: bool):
        """Sets whether notifications should be shown for escape-sequence originated notifications

        :param value: A bool"""
        return self._simple_set("Send Terminal Generated Alerts", value)

    def set_flashing_bell(self, value: bool):
        """Sets whether the bell should flash the screen

        :param value: A bool"""
        return self._simple_set("Flashing Bell", value)

    def set_visual_bell(self, value: bool):
        """Sets whether a bell should be shown when the bell rings

        :param value: A bool"""
        return self._simple_set("Visual Bell", value)

    def set_close_sessions_on_end(self, value: bool):
        """Sets whether the session should close when it ends.

        :param value: A bool"""
        return self._simple_set("Close Sessions On End", value)

    def set_prompt_before_closing(self, value: bool):
        """Sets whether the session should prompt before closign

        :param value: A bool"""
        return self._simple_set("Prompt Before Closing 2", value)

    def set_session_close_undo_timeout(self, value: float):
        """Sets amount of time you can undo closing a session

        :param value: A float at least 0"""
        return self._simple_set("Session Close Undo Timeout", value)

    def set_reduce_flicker(self, value: bool):
        """Sets whether the flicker fixer is on.

        :param value: A bool"""
        return self._simple_set("Reduce Flicker", value)

    def set_send_code_when_idle(self, value: bool):
        """Sets whether to send a code when idle

        :param value: A bool"""
        return self._simple_set("Send Code When Idle", value)

    def set_application_keypad_allowed(self, value: bool):
        """Sets whether the terminal may be placed in application keypad mode

        :param value: A bool"""
        return self._simple_set("Application Keypad Allowed", value)

    def set_place_prompt_at_first_column(self, value: bool):
        """Sets whether the prompt should always begin at the first column (requires shell
        integration)

        :param value: A bool"""
        return self._simple_set("Place Prompt at First Column", value)

    def set_show_mark_indicators(self, value: bool):
        """Sets whether mark indicators should be visible

        :param value: A bool"""
        return self._simple_set("Show Mark Indicators", value)

    def set_idle_code(self, value: int):
        """Sets the ASCII code to send on idle

        :param value: An int in 0...255"""
        return self._simple_set("Idle Code", value)

    def set_idle_period(self, value: float):
        """Sets how often to send a code when idle

        :param value: A float at least 0"""
        return self._simple_set("Idle Period", value)

    def set_unicode_version(self, value: bool):
        """Sets the unicode version for wcwidth

        :param value: A bool"""
        return self._simple_set("Unicode Version", value)

    def set_cursor_type(self, value: CursorType):
        """Sets the cursor type

        :param value: The new value"""
        return self._simple_set("Cursor Type", value)

    def set_thin_strokes(self, value: ThinStrokes):
        """Sets whether thin strokes are used.

        :param value: The new value."""
        return self._simple_set("Thin Strokes", value)

    def set_unicode_normalization(self, value: UnicodeNormalization):
        """Sets the unicode normalization form to use

        :param value: the new value"""
        return self._simple_set("Unicode Normalization", value)

    def set_character_encoding(self, value: CharacterEncoding):
        """Sets the character encoding

        :param value: The new value."""
        return self._simple_set("Character Encoding", value)

    def set_left_option_key_sends(self, value: OptionKeySends):
        """Sets the behavior of the left option key.

        :param value: The new value."""
        return self._simple_set("Option Key Sends", value)

    def set_right_option_key_sends(self, value: OptionKeySends):
        """Sets the behavior of the right option key.

        :param value: The new value."""
        return self._simple_set("Right Option Key Sends", value)

    def set_triggers(self, value: typing.List[typing.Dict[str, typing.Any]]):
        """Sets the triggers.

        :param value: A list of dicts of trigger definitions."""
        return self._simple_set("Triggers", value)

    def set_smart_selection_rules(self, value: typing.List[typing.Dict[str, typing.Any]]):
        """Sets the smart selection rules.

        :param value: A list of dicts of smart selection rules"""
        return self._simple_set("Smart Selection Rules", value)

    def set_semantic_history(self, value: typing.Dict[str, typing.Any]):
        """Sets the semantic history prefs.

        :param value: Semantic history settings dict."""
        return self._simple_set("Semantic History", value)

    def set_automatic_profile_switching_rules(self, value: typing.List[str]):
        """Sets the automatic profile switching rules.

        :param value: A list of rules (strings)."""
        return self._simple_set("Bound Hosts", value)

    def set_advanced_working_directory_window_setting(self, value: InitialWorkingDirectory):
        """Sets the advanced working directory window setting.

        :param value: The new value. Excludes Advanced."""
        return self._simple_set("AWDS Window Option", value)

    def set_advanced_working_directory_window_directory(self, value: str):
        """Sets the advanced working directory window directory.

        :param value: Path."""
        return self._simple_set("AWDS Window Directory", value)

    def set_advanced_working_directory_tab_setting(self, value: InitialWorkingDirectory):
        """Sets the advanced working directory tab setting.

        :param value: The new value. Excludes Advanced."""
        return self._simple_set("AWDS Tab Option", value)

    def set_advanced_working_directory_tab_directory(self, value: str):
        """Sets the advanced working directory tab directory.

        :param value: Path."""
        return self._simple_set("AWDS Tab Directory", value)

    def set_advanced_working_directory_pane_setting(self, value: InitialWorkingDirectory):
        """Sets the advanced working directory pane setting.

        :param value: The new value. Excludes Advanced."""
        return self._simple_set("AWDS Pane Option", value)

    def set_advanced_working_directory_pane_directory(self, value: str):
        """Sets the advanced working directory pane directory.

        :param value: Path."""
        return self._simple_set("AWDS Pane Directory", value)

    def set_normal_font(self, value: str):
        """Sets the normal font.

        The normal font is used for either ASCII or all characters depending on
        whether a separate font is used for non-ascii.

        :param value: Font name and size as a string."""
        return self._simple_set("Normal Font", value)

    def set_non_ascii_font(self, value: str):
        """Sets the non-ASCII font.

        This is used for non-ASCII characters if use_non_ascii_font is enabled.

        :param value: Font name and size as a string."""
        return self._simple_set("Non Ascii Font", value)

    def set_background_image_location(self, value: str):
        """Sets path to the background image.

        :param value: Path."""
        return self._simple_set("Background Image Location", value)

    def set_key_mappings(self, value: typing.Dict[str, str]):
        """Sets the keyboard shortcuts.

        :param value: Dictionary mapping keystroke to action."""
        return self._simple_set("Keyboard Map", value)

    def set_touchbar_mappings(self, value: typing.Dict[str, str]):
        """Sets the touchbar actions.

        :param value: Dictionary mapping touch bar item to action."""
        return self._simple_set("Touch Bar Map", value)

    def set_use_custom_command(self, value: str):
        """"Sets whether to use a custom command when the session is created.

        :param value: The string "Yes" or "No".
        """
        return self._simple_set("Custom Command", value)

    def set_command(self, value: str):
        """"The command to run when the session starts.

        custom_command must be set to "Yes" or this will be ignored.

        :param value: A string giving the command to run.
        """
        return self._simple_set("Command", value)

    def set_initial_directory_mode(self, value: InitialWorkingDirectory):
        """Sets whether to use a custom (not home) initial working directory.

        :param value: The new value.
        """
        return self._simple_set("Custom Directory", value.value)

    def set_custom_directory(self, value: str):
        """Sets the initial working directory.

        The initial_directory_mode must be set to "Yes" for this to take effect.
        """
        return self._simple_set("Working Directory", value)

    def set_icon_mode(self, value: IconMode):
        """Sets the icon mode.

        :param value: The icon mode.
        """
        return self._simple_set("Icon", value.value)

    def set_custom_icon_path(self, value: str):
        """Sets the path of the custom icon.

        The `icon_mode` must be set to `CUSTOM`.
        """
        return self._simple_set("Custom Icon Path", value)

    def set_title_components(self, value: typing.List[TitleComponents]):
        """Sets which components are visible in the session's title, or selects a custom component.

        If it is set to `CUSTOM` then the title_function must be set properly.
        """
        n = 0
        for c in value:
            n += c.value
        return self._simple_set("Title Components", n)

    def set_title_function(self, display_name: str, identifier: str):
        """Sets the function call for the session title provider and its display name for the UI.

        :param display_name: This is shown in the Title Components menu in the UI.
        :identifier: The unique identifier, typically a backwards domain name.

        This takes effect only when the title_components property is set to `CUSTOM`.
        """
        return self._simple_set("Title Function", [display_name, identifier])

    def set_badge_top_margin(self, value: int):
        """Sets the top margin of the badge.

        :param value: The new value in points.
        """
        return self._simple_set("Badge Top Margin", value)

    def set_badge_right_margin(self, value: int):
        """Sets the right margin of the badge.

        :param value: The new value in points.
        """
        return self._simple_set("Badge Right Margin", value)

    def set_badge_max_width(self, value: int):
        """Sets the max width of the badge.

        :param value: The new value in points.
        """
        return self._simple_set("Badge Max Width", value)

    def set_badge_max_height(self, value: int):
        """Sets the max height of the badge.

        :param value: The new value in points.
        """
        return self._simple_set("Badge Max Height", value)


    def set_badge_font(self, value: str):
        """Sets the font of the badge.

        :param value: The new font name, like "Helvetica"
        """
        return self._simple_set("Badge Font", value)

    def set_use_custom_window_title(self, value: bool):
        """Sets whether the custom window title is used.

        :param value: Should the custom window title in the profile be used?
        """
        return self._simple_set("Use Custom Window Title", value)

    def set_custom_window_title(self, value: str):
        """Sets the custom window title.

        This will only be used if use_custom_window_title is True.

        :param value: The new value. An interpolated string.
        """
        return self._simple_set("Custom Window Title", value)

    def set_use_transparency_initially(self, value: bool):
        """Should a window created with this profile respect the transparency setting?

        :param value: If True, use transparency; if False, force the window to be opaque (but it can be toggled with View > Use Transparency).
        """
        return self._simple_set("Initial Use Transparency", value)

    def set_status_bar_enabled(self, value: bool):
        """Should the status bar be enabled?

        :param value: If True, the status bar will be shown.
        """
        return self._simple_set("Show Status Bar", value)

    def set_use_csi_u(self, value: bool):
        """Report keystrokes with CSI u protocol?

        :param value: If True, CSI u will be enabled.
        """
        return self._simple_set("Use libtickit protocol", value)

    def set_triggers_use_interpolated_strings(self, value: bool):
        """Should trigger parameters be interpreted as interpolated strings?
        """
        return self._simple_set("Triggers Use Interpolated Strings", value)


class WriteOnlyProfile:
    """A profile that can be modified but not read. Useful for changing many
    sessions' profiles at once without knowing what they are."""
    def __init__(self, session_id, connection, guid=None):
        assert session_id != "all"
        self.connection = connection
        self.session_id = session_id
        self.__guid = guid

    async def _async_simple_set(self, key: str, value: typing.Any):
        """value is a json type"""
        await iterm2.rpc.async_set_profile_property(
                self.connection,
                self.session_id,
                key,
                value,
                self._guids_for_set())

    async def _async_color_set(self, key, value):
        if value is None:
            await iterm2.rpc.async_set_profile_property(
                self.connection,
                self.session_id,
                key,
                "null",
                self._guids_for_set())
        else:
            await iterm2.rpc.async_set_profile_property(
                self.connection,
                self.session_id,
                key,
                value.get_dict(),
                self._guids_for_set())

    def _guids_for_set(self):
        if self.session_id is None:
            assert self.__guid is not None
            return [self.__guid]
        else:
            return self.session_id

    async def async_set_color_preset(self, preset: iterm2.colorpresets.ColorPreset):
        """Sets the color preset.

        :param preset: The new value.

        .. seealso::
            * Example ":ref:`colorhost_example`"
            * Example ":ref:`random_color_example`"
            * Example ":ref:`theme_example`"
            * Example ":ref:`darknight`"
        """
        coros = []
        for value in preset.values:
            coro = self._async_color_set(
                    value.key,
                    iterm2.color.Color(
                        value.red,
                        value.green,
                        value.blue,
                        value.alpha,
                        value.color_space))
            coros.append(coro)
        await asyncio.gather(*coros)


    async def async_set_foreground_color(self, value: 'iterm2.color.Color'):
        """Sets the foreground color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Foreground Color", value)

    async def async_set_background_color(self, value: 'iterm2.color.Color'):
        """Sets the background color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Background Color", value)

    async def async_set_bold_color(self, value: 'iterm2.color.Color'):
        """Sets the bold text color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Bold Color", value)

    async def async_set_link_color(self, value: 'iterm2.color.Color'):
        """Sets the link color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Link Color", value)

    async def async_set_selection_color(self, value: 'iterm2.color.Color'):
        """Sets the selection background color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Selection Color", value)

    async def async_set_selected_text_color(self, value: 'iterm2.color.Color'):
        """Sets the selection text color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Selected Text Color", value)

    async def async_set_cursor_color(self, value: 'iterm2.color.Color'):
        """Sets the cursor color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Cursor Color", value)

    async def async_set_cursor_text_color(self, value: 'iterm2.color.Color'):
        """Sets the cursor text color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Cursor Text Color", value)

    async def async_set_ansi_0_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 0 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 0 Color", value)

    async def async_set_ansi_1_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 1 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 1 Color", value)

    async def async_set_ansi_2_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 2 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 2 Color", value)

    async def async_set_ansi_3_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 3 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 3 Color", value)

    async def async_set_ansi_4_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 4 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 4 Color", value)

    async def async_set_ansi_5_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 5 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 5 Color", value)

    async def async_set_ansi_6_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 6 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 6 Color", value)

    async def async_set_ansi_7_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 7 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 7 Color", value)

    async def async_set_ansi_8_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 8 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 8 Color", value)

    async def async_set_ansi_9_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 9 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 9 Color", value)

    async def async_set_ansi_10_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 10 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 10 Color", value)

    async def async_set_ansi_11_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 11 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 11 Color", value)

    async def async_set_ansi_12_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 12 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 12 Color", value)

    async def async_set_ansi_13_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 13 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 13 Color", value)

    async def async_set_ansi_14_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 14 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 14 Color", value)

    async def async_set_ansi_15_color(self, value: 'iterm2.color.Color'):
        """Sets the ANSI 15 color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Ansi 15 Color", value)

    async def async_set_smart_cursor_color(self, value: 'iterm2.color.Color'):
        """Sets the smart cursor color.

        :param value: A :class:`Color`"""
        return await self._async_simple_set("Smart Cursor Color", value)

    async def async_set_tab_color(self, value: 'iterm2.color.Color'):
        """Sets the tab color.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Tab Color", value)

    async def async_set_underline_color(self, value: 'iterm2.color.Color'):
        """Sets the underline color.

        :param value: A :class:`Color` or None"""
        return await self._async_color_set("Underline Color", value)

    async def async_set_cursor_guide_color(self, value: 'iterm2.color.Color'):
        """Sets the cursor guide color. The alpha value is respected.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Cursor Guide Color", value)

    async def async_set_badge_color(self, value: 'iterm2.color.Color'):
        """Sets the badge color. The alpha value is respected.

        :param value: A :class:`Color`"""
        return await self._async_color_set("Badge Color", value)

    async def async_set_name(self, value: str):
        """Sets the name.

        :param value: A string"""
        return await self._async_simple_set("Name", value)

    async def async_set_badge_text(self, value: 'iterm2.color.Color'):
        """Sets the badge text.

        :param value: A :class:`Color`"""
        return await self._async_simple_set("Badge Text", value)

    async def async_set_answerback_string(self, value: str):
        """Sets the answerback string.

        :param value: A string"""
        return await self._async_simple_set("Answerback String", value)

    async def async_set_use_cursor_guide(self, value: bool):
        """Sets whether the cursor guide should be used.

        :param value: A boolean"""
        return await self._async_simple_set("Use Cursor Guide", value)

    async def async_set_use_tab_color(self, value: str):
        """Sets whether the tab color should be used.

        :param value: A string"""
        return await self._async_simple_set("Use Tab Color", value)

    async def async_set_use_underline_color(self, value: 'iterm2.color.Color'):
        """Sets the underline color.

        :param value: A :class:`Color`"""
        return await self._async_simple_set("Use Underline Color", value)

    async def async_set_minimum_contrast(self, value: float):
        """Sets the minimum contrast.

        :param value: A float in 0 to 1"""
        return await self._async_simple_set("Minimum Contrast", value)

    async def async_set_cursor_boost(self, value: float):
        """Sets the cursor boost level.

        :param value: A float in 0 to 1"""
        return await self._async_simple_set("Cursor Boost", value)

    async def async_set_blinking_cursor(self, value: bool):
        """Sets whether the cursor blinks.

        :param value: A bool"""
        return await self._async_simple_set("Blinking Cursor", value)

    async def async_set_use_bold_font(self, value: bool):
        """Sets whether to use the bold variant of the font for bold text.

        :param value: A bool"""
        return await self._async_simple_set("Use Bold Font", value)

    async def async_set_ascii_ligatures(self, value: bool):
        """Sets whether ligatures should be used for ASCII text.

        :param value: A bool"""
        return await self._async_simple_set("ASCII Ligatures", value)

    async def async_set_non_ascii_ligatures(self, value: bool):
        """Sets whether ligatures should be used for non-ASCII text.

        :param value: A bool"""
        return await self._async_simple_set("Non-ASCII Ligatures", value)

    async def async_set_use_bright_bold(self, value: bool):
        """Sets whether bright colors should be used for bold text.

        :param value: A bool"""
        return await self._async_simple_set("Use Bright Bold", value)

    async def async_set_blink_allowed(self, value: bool):
        """Sets whether blinking text is allowed.

        :param value: A bool"""
        return await self._async_simple_set("Blink Allowed", value)

    async def async_set_use_italic_font(self, value: bool):
        """Sets whether italic text is allowed.

        :param value: A bool"""
        return await self._async_simple_set("Use Italic Font", value)

    async def async_set_ambiguous_double_width(self, value: bool):
        """Sets whether ambiguous-width text should be treated as double-width.

        :param value: A bool"""
        return await self._async_simple_set("Ambiguous Double Width", value)

    async def async_set_horizontal_spacing(self, value: float):
        """Sets the fraction of horizontal spacing.

        :param value: A float at least 0"""
        return await self._async_simple_set("Horizontal Spacing", value)

    async def async_set_vertical_spacing(self, value: float):
        """Sets the fraction of vertical spacing.

        :param value: A float at least 0"""
        return await self._async_simple_set("Vertical Spacing", value)

    async def async_set_use_non_ascii_font(self, value: bool):
        """Sets whether to use a different font for non-ASCII text.

        :param value: A bool"""
        return await self._async_simple_set("Use Non-ASCII Font", value)

    async def async_set_transparency(self, value: float):
        """Sets the level of transparency.

        :param value: A float between 0 and 1"""
        return await self._async_simple_set("Transparency", value)

    async def async_set_blur(self, value: bool):
        """Sets whether background blur should be enabled.

        :param value: A bool"""
        return await self._async_simple_set("Blur", value)

    async def async_set_blur_radius(self, value: float):
        """Sets the blur radius (how blurry). Requires blur to be enabled.

        :param value: A float between 0 and 30"""
        return await self._async_simple_set("Blur Radius", value)

    async def async_set_background_image_mode(self, value: BackgroundImageMode):
        """Sets how the background image is draw.

        :param value: The new value."""
        return await self._async_simple_set("Background Image Mode", value)

    async def async_set_blend(self, value: float):
        """Sets how much the default background color gets blended with the background image.

        .. seealso:: Example ":ref:`blending_example`"

        :param value: A float in 0 to 1"""
        return await self._async_simple_set("Blend", value)

    async def async_set_sync_title(self, value: bool):
        """Sets whether the profile name stays in the tab title, even if changed by an escape
        sequence.

        :param value: A bool"""
        return await self._async_simple_set("Sync Title", value)

    async def async_set_disable_window_resizing(self, value: bool):
        """Sets whether the terminal can resize the window with an escape sequence.

        :param value: A bool"""
        return await self._async_simple_set("Disable Window Resizing", value)

    async def async_set_only_the_default_bg_color_uses_transparency(self, value: bool):
        """Sets whether window transparency shows through non-default background colors.

        :param value: A bool"""
        return await self._async_simple_set("Only The Default BG Color Uses Transparency", value)

    async def async_set_ascii_anti_aliased(self, value: bool):
        """Sets whether ASCII text is anti-aliased.

        :param value: A bool"""
        return await self._async_simple_set("ASCII Anti Aliased", value)

    async def async_set_non_ascii_anti_aliased(self, value: bool):
        """Sets whether non-ASCII text is anti-aliased.

        :param value: A bool"""
        return await self._async_simple_set("Non-ASCII Anti Aliased", value)

    async def async_set_scrollback_lines(self, value: int):
        """Sets the number of scrollback lines.

        :param value: An int at least 0"""
        return await self._async_simple_set("Scrollback Lines", value)

    async def async_set_unlimited_scrollback(self, value: bool):
        """Sets whether the scrollback buffer's length is unlimited.

        :param value: A bool"""
        return await self._async_simple_set("Unlimited Scrollback", value)

    async def async_set_scrollback_with_status_bar(self, value: bool):
        """Sets whether text gets appended to scrollback when there is an app status bar

        :param value: A bool"""
        return await self._async_simple_set("Scrollback With Status Bar", value)

    async def async_set_scrollback_in_alternate_screen(self, value: bool):
        """Sets whether text gets appended to scrollback in alternate screen mode

        :param value: A bool"""
        return await self._async_simple_set("Scrollback in Alternate Screen", value)

    async def async_set_mouse_reporting(self, value: bool):
        """Sets whether mouse reporting is allowed

        :param value: A bool"""
        return await self._async_simple_set("Mouse Reporting", value)

    async def async_set_mouse_reporting_allow_mouse_wheel(self, value: bool):
        """Sets whether mouse reporting reports the mouse wheel's movements.

        :param value: A bool"""
        return await self._async_simple_set("Mouse Reporting allow mouse wheel", value)

    async def async_set_allow_title_reporting(self, value: bool):
        """Sets whether the session title can be reported

        :param value: A bool"""
        return await self._async_simple_set("Allow Title Reporting", value)

    async def async_set_allow_title_setting(self, value: bool):
        """Sets whether the session title can be changed by escape sequence

        :param value: A bool"""
        return await self._async_simple_set("Allow Title Setting", value)

    async def async_set_disable_printing(self, value: bool):
        """Sets whether printing by escape sequence is disabled.

        :param value: A bool"""
        return await self._async_simple_set("Disable Printing", value)

    async def async_set_disable_smcup_rmcup(self, value: bool):
        """Sets whether alternate screen mode is disabled

        :param value: A bool"""
        return await self._async_simple_set("Disable Smcup Rmcup", value)

    async def async_set_silence_bell(self, value: bool):
        """Sets whether the bell makes noise.

        :param value: A bool"""
        return await self._async_simple_set("Silence Bell", value)

    async def async_set_bm_growl(self, value: bool):
        """Sets whether notifications should be shown.

        :param value: A bool"""
        return await self._async_simple_set("BM Growl", value)

    async def async_set_send_bell_alert(self, value: bool):
        """Sets whether notifications should be shown for the bell ringing

        :param value: A bool"""
        return await self._async_simple_set("Send Bell Alert", value)

    async def async_set_send_idle_alert(self, value: bool):
        """Sets whether notifications should be shown for becoming idle

        :param value: A bool"""
        return await self._async_simple_set("Send Idle Alert", value)

    async def async_set_send_new_output_alert(self, value: bool):
        """Sets whether notifications should be shown for new output

        :param value: A bool"""
        return await self._async_simple_set("Send New Output Alert", value)

    async def async_set_send_session_ended_alert(self, value: bool):
        """Sets whether notifications should be shown for a session ending

        :param value: A bool"""
        return await self._async_simple_set("Send Session Ended Alert", value)

    async def async_set_send_terminal_generated_alerts(self, value: bool):
        """Sets whether notifications should be shown for escape-sequence originated notifications

        :param value: A bool"""
        return await self._async_simple_set("Send Terminal Generated Alerts", value)

    async def async_set_flashing_bell(self, value: bool):
        """Sets whether the bell should flash the screen

        :param value: A bool"""
        return await self._async_simple_set("Flashing Bell", value)

    async def async_set_visual_bell(self, value: bool):
        """Sets whether a bell should be shown when the bell rings

        :param value: A bool"""
        return await self._async_simple_set("Visual Bell", value)

    async def async_set_close_sessions_on_end(self, value: bool):
        """Sets whether the session should close when it ends.

        :param value: A bool"""
        return await self._async_simple_set("Close Sessions On End", value)

    async def async_set_prompt_before_closing(self, value: bool):
        """Sets whether the session should prompt before closign

        :param value: A bool"""
        return await self._async_simple_set("Prompt Before Closing 2", value)

    async def async_set_session_close_undo_timeout(self, value: float):
        """Sets amount of time you can undo closing a session

        :param value: A float at least 0"""
        return await self._async_simple_set("Session Close Undo Timeout", value)

    async def async_set_reduce_flicker(self, value: bool):
        """Sets whether the flicker fixer is on.

        :param value: A bool"""
        return await self._async_simple_set("Reduce Flicker", value)

    async def async_set_send_code_when_idle(self, value: bool):
        """Sets whether to send a code when idle

        :param value: A bool"""
        return await self._async_simple_set("Send Code When Idle", value)

    async def async_set_application_keypad_allowed(self, value: bool):
        """Sets whether the terminal may be placed in application keypad mode

        :param value: A bool"""
        return await self._async_simple_set("Application Keypad Allowed", value)

    async def async_set_place_prompt_at_first_column(self, value: bool):
        """Sets whether the prompt should always begin at the first column (requires shell
        integration)

        :param value: A bool"""
        return await self._async_simple_set("Place Prompt at First Column", value)

    async def async_set_show_mark_indicators(self, value: bool):
        """Sets whether mark indicators should be visible

        :param value: A bool"""
        return await self._async_simple_set("Show Mark Indicators", value)

    async def async_set_idle_code(self, value: int):
        """Sets the ASCII code to send on idle

        :param value: An int in 0...255"""
        return await self._async_simple_set("Idle Code", value)

    async def async_set_idle_period(self, value: float):
        """Sets how often to send a code when idle

        :param value: A float at least 0"""
        return await self._async_simple_set("Idle Period", value)

    async def async_set_unicode_version(self, value: bool):
        """Sets the unicode version for wcwidth

        :param value: A bool"""
        return await self._async_simple_set("Unicode Version", value)

    async def async_set_cursor_type(self, value: CursorType):
        """Sets the cursor type

        :param value: The new value."""
        return await self._async_simple_set("Cursor Type", value)

    async def async_set_thin_strokes(self, value: ThinStrokes):
        """Sets whether thin strokes are used.

        :param value: The new value."""
        return await self._async_simple_set("Thin Strokes", value)

    async def async_set_unicode_normalization(self, value: UnicodeNormalization):
        """Sets the unicode normalization form to use

        :param value: The new value."""
        return await self._async_simple_set("Unicode Normalization", value)

    async def async_set_character_encoding(self, value: CharacterEncoding):
        """Sets the character encoding

        :param value: The new value."""
        return await self._async_simple_set("Character Encoding", value)

    async def async_set_left_option_key_sends(self, value: OptionKeySends):
        """Sets the behavior of the left option key.

        :param value: The new value."""
        return await self._async_simple_set("Option Key Sends", value)

    async def async_set_right_option_key_sends(self, value: OptionKeySends):
        """Sets the behavior of the right option key.

        :param value: The new value."""
        return await self._async_simple_set("Right Option Key Sends", value)

    async def async_set_triggers(self, value: typing.List[typing.Dict[str, typing.Any]]):
        """Sets the triggers.

        :param value: A list of dicts of trigger definitions."""
        return await self._async_simple_set("Triggers", value)

    async def async_set_smart_selection_rules(self, value: typing.List[typing.Dict[str, typing.Any]]):
        """Sets the smart selection rules.

        :param value: A list of dicts of smart selection rules"""
        return await self._async_simple_set("Smart Selection Rules", value)

    async def async_set_semantic_history(self, value: typing.Dict[str, typing.Any]):
        """Sets the semantic history prefs.

        :param value: Semantic history settings dict."""
        return await self._async_simple_set("Semantic History", value)

    async def async_set_automatic_profile_switching_rules(self, value: typing.List[str]):
        """Sets the automatic profile switching rules.

        :param value: A list of rules (strings)."""
        return await self._async_simple_set("Bound Hosts", value)

    async def async_set_advanced_working_directory_window_setting(self, value: InitialWorkingDirectory):
        """Sets the advanced working directory window setting.

        :param value: New value. Excludes Advanced."""
        return await self._async_simple_set("AWDS Window Option", value)

    async def async_set_advanced_working_directory_window_directory(self, value: str):
        """Sets the advanced working directory window directory.

        :param value: Path."""
        return await self._async_simple_set("AWDS Window Directory", value)

    async def async_set_advanced_working_directory_tab_setting(self, value: InitialWorkingDirectory):
        """Sets the advanced working directory tab setting.

        :param value: New value. Excludes Advanced."""
        return await self._async_simple_set("AWDS Tab Option", value)

    async def async_set_advanced_working_directory_tab_directory(self, value: str):
        """Sets the advanced working directory tab directory.

        :param value: Path."""
        return await self._async_simple_set("AWDS Tab Directory", value)

    async def async_set_advanced_working_directory_pane_setting(self, value: InitialWorkingDirectory):
        """Sets the advanced working directory pane setting.

        :param value: New value. Excludes Advanced."""
        return await self._async_simple_set("AWDS Pane Option", value)

    async def async_set_advanced_working_directory_pane_directory(self, value: str):
        """Sets the advanced working directory pane directory.

        :param value: Path."""
        return await self._async_simple_set("AWDS Pane Directory", value)

    async def async_set_normal_font(self, value: str):
        """Sets the normal font.

        The normal font is used for either ASCII or all characters depending on
        whether a separate font is used for non-ascii.

        :param value: Font name and size as a string."""
        return await self._async_simple_set("Normal Font", value)

    async def async_set_non_ascii_font(self, value: str):
        """Sets the non-ASCII font.

        This is used for non-ASCII characters if use_non_ascii_font is enabled.

        :param value: Font name and size as a string."""
        return await self._async_simple_set("Non Ascii Font", value)

    async def async_set_background_image_location(self, value: str):
        """Sets path to the background image.

        :param value: Path."""
        return await self._async_simple_set("Background Image Location", value)

    async def async_set_key_mappings(self, value: typing.Dict[str, typing.Any]):
        """Sets the keyboard shortcuts.

        :param value: Dictionary mapping keystroke to action."""
        return await self._async_simple_set("Keyboard Map", value)

    async def async_set_touchbar_mappings(self, value: typing.Dict[str, typing.Any]):
        """Sets the touchbar actions.

        :param value: Dictionary mapping touch bar item to action."""
        return await self._async_simple_set("Touch Bar Map", value)

    async def async_set_use_custom_command(self, value: str):
        """"Sets whether to use a custom command when the session is created.

        :param value: The string "Yes" or "No".
        """
        return await self._async_simple_set("Custom Command", value)

    async def async_set_command(self, value: str):
        """"The command to run when the session starts.

        custom_command must be set to "Yes" or this will be ignored.

        :param value: A string giving the command to run.
        """
        return await self._async_simple_set("Command", value)

    async def async_set_initial_directory_mode(self, value: InitialWorkingDirectory):
        """Sets whether to use a custom (not home) initial working directory.

        :param value: The new value.
        """
        return await self._async_simple_set("Custom Directory", value.value)

    async def async_set_custom_directory(self, value: str):
        """Sets the initial working directory.

        The initial_directory_mode must be set to "Yes" for this to take effect.

        :param value: The path to use.
        """
        return await self._async_simple_set("Working Directory", value)

    async def async_set_icon_mode(self, value: IconMode):
        """Sets the icon mode.

        :param value: The icon mode.
        """
        return await self._async_simple_set("Icon", value.value)

    async def async_set_custom_icon_path(self, value: str):
        """Sets the path of the custom icon.

        The `icon_mode` must be set to `CUSTOM`.
        """
        return await self._async_simple_set("Custom Icon Path", value)

    async def async_set_title_components(self, value: typing.List[TitleComponents]):
        """Sets which components are visible in the session's title, or selects a custom component.

        If it is set to `CUSTOM` then the title_function must be set properly.
        """
        n = 0
        for c in value:
            n += c.value
        return await self._async_simple_set("Title Components", n)

    async def async_set_title_function(self, display_name: str, identifier: str):
        """Sets the function call for the session title provider and its display name for the UI.

        :param display_name: This is shown in the Title Components menu in the UI.
        :identifier: The unique identifier, typically a backwards domain name.

        This takes effect only when the title_components property is set to `CUSTOM`.
        """
        return await self._async_simple_set("Title Function", [display_name, identifier])

    async def async_set_badge_top_margin(self, value: int):
        """Sets the top margin of the badge.

        :param value: The new value in points.
        """
        return await self._async_simple_set("Badge Top Margin", value)

    async def async_set_badge_right_margin(self, value: int):
        """Sets the right margin of the badge.

        :param value: The new value in points.
        """
        return await self._async_simple_set("Badge Right Margin", value)

    async def async_set_badge_max_width(self, value: int):
        """Sets the max width of the badge.

        :param value: The new value in points.
        """
        return await self._async_simple_set("Badge Max Width", value)

    async def async_set_badge_max_height(self, value: int):
        """Sets the max height of the badge.

        :param value: The new value in points.
        """
        return await self._async_simple_set("Badge Max Height", value)


    async def async_set_badge_font(self, value: str):
        """Sets the font of the badge.

        :param value: The new font name, like "Helvetica"
        """
        return await self._async_simple_set("Badge Font", value)

    async def async_set_use_custom_window_title(self, value: bool):
        """Sets whether the custom window title is used.

        :param value: Should the custom window title in the profile be used?
        """
        return await self._async_simple_set("Use Custom Window Title", value)

    async def async_set_custom_window_title(self, value: str):
        """Sets the custom window title.

        This will only be used if use_custom_window_title is True.

        :param value: The new value. An interpolated string.
        """
        return await self._async_simple_set("Custom Window Title", value)

    async def async_set_use_transparency_initially(self, value: bool):
        """Should a window created with this profile respect the transparency setting?

        :param value: If True, use transparency; if False, force the window to be opaque (but it can be toggled with View > Use Transparency).
        """
        return await self._async_simple_set("Initial Use Transparency", value)

    async def async_set_status_bar_enabled(self, value: bool):
        """Should the status bar be enabled?

        :param value: If True, the status bar will be shown.
        """
        return await self._async_simple_set("Show Status Bar", value)

    async def async_set_use_csi_u(self, value: bool):
        """Report keystrokes with CSI u protocol?

        :param value: If True, CSI u will be enabled.
        """
        return await self._async_simple_set("Use libtickit protocol", value)

    async def async_set_triggers_use_interpolated_strings(self, value: bool):
        """Should trigger parameters be interpreted as interpolated strings?
        """
        return await self._async_simple_set("Triggers Use Interpolated Strings", value)

class Profile(WriteOnlyProfile):
    """Represents a profile.

    If a session_id is set then this is the profile attached to a session.
    Otherwise, it is a shared profile."""

    USE_CUSTOM_COMMAND_ENABLED = "Yes"
    USE_CUSTOM_COMMAND_DISABLED = "No"

    @staticmethod
    async def async_get(connection, guids=None) -> typing.List['Profile']:
        """Fetches all profiles with the specified GUIDs.

        :param guids: The profiles to get, or if `None` then all will be returned.

        :returns: A list of :class:`Profile` objects.
        """
        response = await iterm2.rpc.async_list_profiles(connection, guids, None)
        profiles = []
        for responseProfile in response.list_profiles_response.profiles:
            profile = Profile(None, connection, responseProfile.properties)
            profiles.append(profile)
        return profiles


    def __init__(self, session_id, connection, profile_property_list):
        props = {}
        for prop in profile_property_list:
            props[prop.key] = json.loads(prop.json_value)

        guid_key = "Guid"
        if guid_key in props:
            guid = props[guid_key]
        else:
            guid = None

        super().__init__(session_id, connection, guid)

        self.connection = connection
        self.session_id = session_id
        self.__props = props

    def _simple_get(self, key):
        if key in self.__props:
            return self.__props[key]
        else:
            return None

    def _get_optional_bool(self, key):
        if key not in self.__props:
            return None
        return bool(self.__props[key])

    def get_color_with_key(self, key):
        """Returns the color for the request key, or None.

        :param key: A string describing the color. Corresponds to the keys in :class:`~iterm2.ColorPreset.Color`.

        :returns: Either a :class:`~iterm2.color.Color` or `None`.
        """
        try:
            color = iterm2.color.Color()
            color.from_dict(self.__props[key])
            return color
        except ValueError:
            return None
        except KeyError:
            return None

    @property
    def all_properties(self):
        return dict(self.__props)

    @property
    def foreground_color(self):
        """Returns the foreground color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Foreground Color")

    @property
    def background_color(self):
        """Returns the background color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Background Color")

    @property
    def bold_color(self):
        """Returns the bold text color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Bold Color")

    @property
    def link_color(self):
        """Returns the link color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Link Color")

    @property
    def selection_color(self):
        """Returns the selection background color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Selection Color")

    @property
    def selected_text_color(self):
        """Returns the selection text color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Selected Text Color")

    @property
    def cursor_color(self):
        """Returns the cursor color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Cursor Color")

    @property
    def cursor_text_color(self):
        """Returns the cursor text color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Cursor Text Color")

    @property
    def ansi_0_color(self):
        """Returns the ANSI 0 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 0 Color")

    @property
    def ansi_1_color(self):
        """Returns the ANSI 1 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 1 Color")

    @property
    def ansi_2_color(self):
        """Returns the ANSI 2 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 2 Color")

    @property
    def ansi_3_color(self):
        """Returns the ANSI 3 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 3 Color")

    @property
    def ansi_4_color(self):
        """Returns the ANSI 4 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 4 Color")

    @property
    def ansi_5_color(self):
        """Returns the ANSI 5 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 5 Color")

    @property
    def ansi_6_color(self):
        """Returns the ANSI 6 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 6 Color")

    @property
    def ansi_7_color(self):
        """Returns the ANSI 7 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 7 Color")

    @property
    def ansi_8_color(self):
        """Returns the ANSI 8 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 8 Color")

    @property
    def ansi_9_color(self):
        """Returns the ANSI 9 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 9 Color")

    @property
    def ansi_10_color(self):
        """Returns the ANSI 10 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 10 Color")

    @property
    def ansi_11_color(self):
        """Returns the ANSI 11 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 11 Color")

    @property
    def ansi_12_color(self):
        """Returns the ANSI 12 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 12 Color")

    @property
    def ansi_13_color(self):
        """Returns the ANSI 13 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 13 Color")

    @property
    def ansi_14_color(self):
        """Returns the ANSI 14 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 14 Color")

    @property
    def ansi_15_color(self):
        """Returns the ANSI 15 color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Ansi 15 Color")

    @property
    def smart_cursor_color(self):
        """Returns whether smart cursor color is in use for box cursors.

        :returns: A :class:`Color`"""
        return self._simple_get("Smart Cursor Color")

    @property
    def tab_color(self):
        """Returns the tab color.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Tab Color")

    @property
    def underline_color(self):
        """Returns the underline color.

        :returns: A :class:`Color` or None"""
        return self.get_color_with_key("Underline Color")

    @property
    def cursor_guide_color(self):
        """Returns the cursor guide color. The alpha value is respected.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Cursor Guide Color")

    @property
    def badge_color(self):
        """Returns the badge color. The alpha value is respected.

        :returns: A :class:`Color`"""
        return self.get_color_with_key("Badge Color")

    @property
    def name(self):
        """Returns the name.

        :returns: A string"""
        return self._simple_get("Name")

    @property
    def badge_text(self):
        """Returns the badge text.

        :returns: A :class:`Color`"""
        return self._simple_get("Badge Text")

    @property
    def answerback_string(self):
        """Returns the answerback string.

        :returns: A string"""
        return self._simple_get("Answerback String")

    @property
    def use_cursor_guide(self):
        """Returns whether the cursor guide should be used.

        :returns: A boolean"""
        return self._simple_get("Use Cursor Guide")

    @property
    def use_tab_color(self):
        """Returns whether the tab color should be used.

        :returns: A string"""
        return self._simple_get("Use Tab Color")

    @property
    def use_underline_color(self):
        """Returns the underline color.

        :returns: A :class:`Color`"""
        return self._simple_get("Use Underline Color")

    @property
    def minimum_contrast(self):
        """Returns the minimum contrast.

        :returns: A float in 0 to 1"""
        return self._simple_get("Minimum Contrast")

    @property
    def cursor_boost(self):
        """Returns the cursor boost level.

        :returns: A float in 0 to 1"""
        return self._simple_get("Cursor Boost")

    @property
    def blinking_cursor(self):
        """Returns whether the cursor blinks.

        :returns: A bool"""
        return self._simple_get("Blinking Cursor")

    @property
    def use_bold_font(self):
        """Returns whether to use the bold variant of the font for bold text.

        :returns: A bool"""
        return self._simple_get("Use Bold Font")

    @property
    def ascii_ligatures(self):
        """Returns whether ligatures should be used for ASCII text.

        :returns: A bool"""
        return self._simple_get("ASCII Ligatures")

    @property
    def non_ascii_ligatures(self):
        """Returns whether ligatures should be used for non-ASCII text.

        :returns: A bool"""
        return self._simple_get("Non-ASCII Ligatures")

    @property
    def use_bright_bold(self):
        """Returns whether bright colors should be used for bold text.

        :returns: A bool"""
        return self._simple_get("Use Bright Bold")

    @property
    def blink_allowed(self):
        """Returns whether blinking text is allowed.

        :returns: A bool"""
        return self._simple_get("Blink Allowed")

    @property
    def use_italic_font(self):
        """Returns whether italic text is allowed.

        :returns: A bool"""
        return self._simple_get("Use Italic Font")

    @property
    def ambiguous_double_width(self):
        """Returns whether ambiguous-width text should be treated as double-width.

        :returns: A bool"""
        return self._simple_get("Ambiguous Double Width")

    @property
    def horizontal_spacing(self):
        """Returns the fraction of horizontal spacing.

        :returns: A float at least 0"""
        return self._simple_get("Horizontal Spacing")

    @property
    def vertical_spacing(self):
        """Returns the fraction of vertical spacing.

        :returns: A float at least 0"""
        return self._simple_get("Vertical Spacing")

    @property
    def use_non_ascii_font(self):
        """Returns whether to use a different font for non-ASCII text.

        :returns: A bool"""
        return self._simple_get("Use Non-ASCII Font")

    @property
    def transparency(self):
        """Returns the level of transparency.

        :returns: A float between 0 and 1"""
        return self._simple_get("Transparency")

    @property
    def blur(self):
        """Returns whether background blur should be enabled.

        :returns: A bool"""
        return self._simple_get("Blur")

    @property
    def blur_radius(self):
        """Returns the blur radius (how blurry). Requires blur to be enabled.

        :returns: A float between 0 and 30"""
        return self._simple_get("Blur Radius")

    @property
    def background_image_mode(self):
        """Returns how the background image is drawn

        :returns: A `BackgroundImageMode`"""
        return BackgroundImageMode(self._simple_get("Background Image Mode"))

    @property
    def blend(self):
        """Returns tow much the default background color gets blended with the background image.

        :returns: A float in 0 to 1"""
        return self._simple_get("Blend")

    @property
    def sync_title(self):
        """Returns whether the profile name stays in the tab title, even if changed by an escape
        sequence.

        :returns: A bool"""
        return self._simple_get("Sync Title")

    @property
    def disable_window_resizing(self):
        """Returns whether the terminal can resize the window with an escape sequence.

        :returns: A bool"""
        return self._simple_get("Disable Window Resizing")

    @property
    def only_the_default_bg_color_uses_transparency(self):
        """Returns whether window transparency shows through non-default background colors.

        :returns: A bool"""
        return self._simple_get("Only The Default BG Color Uses Transparency")

    @property
    def ascii_anti_aliased(self):
        """Returns whether ASCII text is anti-aliased.

        :returns: A bool"""
        return self._simple_get("ASCII Anti Aliased")

    @property
    def non_ascii_anti_aliased(self):
        """Returns whether non-ASCII text is anti-aliased.

        :returns: A bool"""
        return self._simple_get("Non-ASCII Anti Aliased")

    @property
    def scrollback_lines(self):
        """Returns the number of scrollback lines.

        :returns: An int at least 0"""
        return self._simple_get("Scrollback Lines")

    @property
    def unlimited_scrollback(self):
        """Returns whether the scrollback buffer's length is unlimited.

        :returns: A bool"""
        return self._simple_get("Unlimited Scrollback")

    @property
    def scrollback_with_status_bar(self):
        """Returns whether text gets appended to scrollback when there is an app status bar

        :returns: A bool"""
        return self._simple_get("Scrollback With Status Bar")

    @property
    def scrollback_in_alternate_screen(self):
        """Returns whether text gets appended to scrollback in alternate screen mode

        :returns: A bool"""
        return self._simple_get("Scrollback in Alternate Screen")

    @property
    def mouse_reporting(self):
        """Returns whether mouse reporting is allowed

        :returns: A bool"""
        return self._simple_get("Mouse Reporting")

    @property
    def mouse_reporting_allow_mouse_wheel(self):
        """Returns whether mouse reporting reports the mouse wheel's movements.

        :returns: A bool"""
        return self._simple_get("Mouse Reporting allow mouse wheel")

    @property
    def allow_title_reporting(self):
        """Returns whether the session title can be reported

        :returns: A bool"""
        return self._simple_get("Allow Title Reporting")

    @property
    def allow_title_setting(self):
        """Returns whether the session title can be changed by escape sequence

        :returns: A bool"""
        return self._simple_get("Allow Title Setting")

    @property
    def disable_printing(self):
        """Returns whether printing by escape sequence is disabled.

        :returns: A bool"""
        return self._simple_get("Disable Printing")

    @property
    def disable_smcup_rmcup(self):
        """Returns whether alternate screen mode is disabled

        :returns: A bool"""
        return self._simple_get("Disable Smcup Rmcup")

    @property
    def silence_bell(self):
        """Returns whether the bell makes noise.

        :returns: A bool"""
        return self._simple_get("Silence Bell")

    @property
    def bm_growl(self):
        """Returns whether notifications should be shown.

        :returns: A bool"""
        return self._simple_get("BM Growl")

    @property
    def send_bell_alert(self):
        """Returns whether notifications should be shown for the bell ringing

        :returns: A bool"""
        return self._simple_get("Send Bell Alert")

    @property
    def send_idle_alert(self):
        """Returns whether notifications should be shown for becoming idle

        :returns: A bool"""
        return self._simple_get("Send Idle Alert")

    @property
    def send_new_output_alert(self):
        """Returns whether notifications should be shown for new output

        :returns: A bool"""
        return self._simple_get("Send New Output Alert")

    @property
    def send_session_ended_alert(self):
        """Returns whether notifications should be shown for a session ending

        :returns: A bool"""
        return self._simple_get("Send Session Ended Alert")

    @property
    def send_terminal_generated_alerts(self):
        """Returns whether notifications should be shown for escape-sequence originated
        notifications

        :returns: A bool"""
        return self._simple_get("Send Terminal Generated Alerts")

    @property
    def flashing_bell(self):
        """Returns whether the bell should flash the screen

        :returns: A bool"""
        return self._simple_get("Flashing Bell")

    @property
    def visual_bell(self):
        """Returns whether a bell should be shown when the bell rings

        :returns: A bool"""
        return self._simple_get("Visual Bell")

    @property
    def close_sessions_on_end(self):
        """Returns whether the session should close when it ends.

        :returns: A bool"""
        return self._simple_get("Close Sessions On End")

    @property
    def prompt_before_closing(self):
        """Returns whether the session should prompt before closign

        :returns: A bool"""
        return self._simple_get("Prompt Before Closing 2")

    @property
    def session_close_undo_timeout(self):
        """Returns tmount of time you can undo closing a session

        :returns: A float at least 0"""
        return self._simple_get("Session Close Undo Timeout")

    @property
    def reduce_flicker(self):
        """Returns whether the flicker fixer is on.

        :returns: A bool"""
        return self._simple_get("Reduce Flicker")

    @property
    def send_code_when_idle(self):
        """Returns whether to send a code when idle

        :returns: A bool"""
        return self._simple_get("Send Code When Idle")

    @property
    def application_keypad_allowed(self):
        """Returns whether the terminal may be placed in application keypad mode

        :returns: A bool"""
        return self._simple_get("Application Keypad Allowed")

    @property
    def place_prompt_at_first_column(self):
        """Returns whether the prompt should always begin at the first column (requires shell
        integration)

        :returns: A bool"""
        return self._simple_get("Place Prompt at First Column")

    @property
    def show_mark_indicators(self):
        """Returns whether mark indicators should be visible

        :returns: A bool"""
        return self._simple_get("Show Mark Indicators")

    @property
    def idle_code(self):
        """Returns the ASCII code to send on idle

        :returns: An int in 0...255"""
        return self._simple_get("Idle Code")

    @property
    def idle_period(self):
        """Returns how often to send a code when idle

        :returns: A float at least 0"""
        return self._simple_get("Idle Period")

    @property
    def unicode_version(self):
        """Returns the unicode version for wcwidth

        :returns: A bool"""
        return self._simple_get("Unicode Version")

    @property
    def cursor_type(self) -> CursorType:
        """Returns the cursor type."""
        return self._simple_get("Cursor Type")

    @property
    def thin_strokes(self) -> ThinStrokes:
        """Returns whether thin strokes are used.

        :returns: THIN_STROKES_SETTING_xxx"""
        return self._simple_get("Thin Strokes")

    @property
    def unicode_normalization(self):
        """Returns the unicode normalization form to use

        :returns: UNICODE_NORMALIZATION_xxx"""
        return self._simple_get("Unicode Normalization")

    @property
    def character_encoding(self):
        """Returns the character encoding

        :returns: CHARACTER_ENCODING_xxx"""
        return self._simple_get("Character Encoding")

    @property
    def left_option_key_sends(self):
        """Returns the behavior of the left option key.

        :returns: OPTION_KEY_xxx"""
        return self._simple_get("Option Key Sends")

    @property
    def right_option_key_sends(self):
        """Returns the behavior of the right option key.

        :returns: OPTION_KEY_xxx"""
        return self._simple_get("Right Option Key Sends")

    @property
    def guid(self):
        """Returns globally unique ID for this profile.

        :returns: A string identifying this profile"""
        return self._simple_get("Guid")

    @property
    def triggers(self) -> typing.List[typing.Dict[str, typing.Any]]:
        """The triggers.

        :returns: A list of dicts of trigger definitions."""
        return self._simple_get("Triggers")

    @property
    def smart_selection_rules(self) -> typing.List[typing.Dict[str, typing.Any]]:
        """The smart selection rules.

        :returns: A list of dicts of smart selection rules"""
        return self._simple_get("Smart Selection Rules")

    @property
    def semantic_history(self) -> typing.Dict[str, typing.Any]:
        """The semantic history prefs.

        :returns: Semantic history settings dict."""
        return self._simple_get("Semantic History")

    @property
    def automatic_profile_switching_rules(self) -> typing.List[str]:
        """The automatic profile switching rules.

        :returns: A list of rules (strings)."""
        return self._simple_get("Bound Hosts")

    @property
    def advanced_working_directory_window_setting(self):
        """The advanced working directory window setting.

        :returns: INITIAL_WORKING_DIRECTORY_xxx, excluding ADVANCED."""
        return self._simple_get("AWDS Window Option")

    @property
    def advanced_working_directory_window_directory(self):
        """The advanced working directory window directory.

        :returns: Path."""
        return self._simple_get("AWDS Window Directory")

    @property
    def advanced_working_directory_tab_setting(self):
        """The advanced working directory tab setting.

        :returns: INITIAL_WORKING_DIRECTORY_xxx, excluding ADVANCED."""
        return self._simple_get("AWDS Tab Option")

    @property
    def advanced_working_directory_tab_directory(self):
        """The advanced working directory tab directory.

        :returns: Path."""
        return self._simple_get("AWDS Tab Directory")

    @property
    def advanced_working_directory_pane_setting(self):
        """The advanced working directory pane setting.

        :returns: INITIAL_WORKING_DIRECTORY_xxx, excluding ADVANCED."""
        return self._simple_get("AWDS Pane Option")

    @property
    def advanced_working_directory_pane_directory(self):
        """The advanced working directory pane directory.

        :returns: Path."""
        return self._simple_get("AWDS Pane Directory")

    @property
    def normal_font(self):
        """The normal font.

        The normal font is used for either ASCII or all characters depending on
        whether a separate font is used for non-ascii.

        :returns: Font name and size as a string."""
        return self._simple_get("Normal Font")

    @property
    def non_ascii_font(self):
        """The non-ASCII font.

        This is used for non-ASCII characters if use_non_ascii_font is enabled.

        :returns: Font name and size as a string."""
        return self._simple_get("Non Ascii Font")

    @property
    def background_image_location(self):
        """Gets path to the background image.

        :returns: Path."""
        return self._simple_get("Background Image Location")

    @property
    def key_mappings(self):
        """The keyboard shortcuts.

        :returns: Dictionary mapping keystroke to action."""
        return self._simple_get("Keyboard Map")

    @property
    def touchbar_mappings(self):
        """The touchbar actions.

        :returns: Dictionary mapping touch bar item to action."""
        return self._simple_get("Touch Bar Map")

    @property
    def original_guid(self):
        """The GUID of the original profile from which this one was derived.

        Used for sessions whose profile has been modified from the underlying
        profile. Otherwise not set.

        :returns: Guid"""
        return self._simple_get("Original Guid")

    @property
    def dynamic_profile_parent_name(self):
        """If the profile is a dynamic profile, returns the name of the parent profile.

        :returns: String name"""
        return self._simple_get("Dynamic Profile Parent Name")

    @property
    def dynamic_profile_file_name(self):
        """If the profile is a dynamic profile, returns the path to the file
        from which it came.

        :returns: String file name"""
        return self._simple_get("Dynamic Profile Filename")

    @property
    def use_custom_command(self):
        """"Returns whether to use a custom command when the session is created.

        :returns: Boolean, whether to use a custom command.
        """
        return self._simple_get("Custom Command")

    @property
    def command(self):
        """"The command to run when the session starts.

        :returns: The command to run, provided `use_custom_command` is `True`.
        """
        return self._simple_get("Command")

    @property
    def initial_directory_mode(self):
        """Returns wether to use a custom (not home) initial working directory.

        :returns: "Yes" to use the `custom_directory`. "No" to use the home directory. "Recycle" to reuse the current directory. "Advanced" to respect advanced working directory settings.
        """
        return self._simple_get("Custom Directory")

    @property
    def custom_directory(self):
        """Returns the initial working directory.

        The initial_directory_mode must be set to "Yes" for this to take effect.

        :returns: The specific directory this profile has been set to start in.
        """
        return self._simple_get("Working Directory")

    @property
    def icon_mode(self) -> IconMode:
        """Returns what kind of icon the session shows.

        :returns: The icon mode.
        """
        return self._simple_get("Icon")

    @property
    def custom_icon_path(self) -> typing.Optional[str]:
        """Returns the path of the custom icon.

        The `icon_mode` must be set to `CUSTOM`.
        """
        return self._simple_get("Custom Icon Path")

    @property
    def title_components(self) -> typing.Optional[typing.List[TitleComponents]]:
        """Returns which components are visible in the session's title, or selects a custom component.

        If it is set to `CUSTOM` then the title_function must be set properly.
        """
        l = []
        n = 1
        value = self._simple_get("Title Components")
        while n <= value:
            if (n & value):
                l.append(TitleComponents(n))
            n *= 2
        return l

    @property
    def title_function(self) -> typing.Optional[typing.Tuple[str, str]]:
        """Returns the function call for the session title provider and its display name for the UI.

        :returns: (display name, unique identifier)
        """
        list = self._simple_get("Title Function")
        return (list[0], list[1])

    @property
    def badge_top_margin(self) -> typing.Optional[int]:
        """Returns the top margin of the badge.

        :returns: The new value in points.
        """
        return self._simple_get("Badge Top Margin")

    @property
    def badge_right_margin(self) -> typing.Optional[int]:
        """Returns the right margin of the badge.

        :returns: The new value in points.
        """
        return self._simple_get("Badge Right Margin")

    @property
    def badge_max_width(self) -> typing.Optional[int]:
        """Returns the max width of the badge.

        :returns: The new value in points.
        """
        return self._simple_get("Badge Max Width")

    @property
    def badge_max_height(self) -> typing.Optional[int]:
        """Returns the max height of the badge.

        :returns: The new value in points.
        """
        return self._simple_get("Badge Max Height")


    @property
    def badge_font(self) -> typing.Optional[str]:
        """Returns the font of the badge.

        :returns: The new font name, like "Helvetica"
        """
        return self._simple_get("Badge Font")

    @property
    def use_custom_window_title(self) -> typing.Optional[bool]:
        """Returns whether the custom window title is used.

        :returns: Should the custom window title in the profile be used?
        """
        return self._get_optional_bool("Use Custom Window Title")

    @property
    def custom_window_title(self) -> typing.Optional[str]:
        """Returns the custom window title.

        This will only be used if use_custom_window_title is True.

        :returns: The new value. An interpolated string.
        """
        return self._simple_get("Custom Window Title")

    @property
    def use_transparency_initially(self) -> typing.Optional[bool]:
        """Returns whether a window created with this profile should respect the transparency setting?

        :returns: If True, use transparency; if False, force the window to be opaque (but it can be toggled with View > Use Transparency).
        """
        return self._get_optional_bool("Initial Use Transparency")

    @property
    def status_bar_enabled(self) -> typing.Optional[bool]:
        """Returns whether the status bar should be enabled?

        :returns: If True, the status bar will be shown.
        """
        return self._get_optional_bool("Show Status Bar")

    @property
    def use_csi_u(self) -> typing.Optional[bool]:
        """REturns wehtehr keystrokes will be reported with CSI u protocol

        :returns: If True, CSI u will be enabled.
        """
        return self._get_optional_bool("Use libtickit protocol")

    @property
    def triggers_use_interpolated_strings(self) -> typing.Optional[bool]:
        """Returns whether trigger parameters are interpreted as interpolated strings?
        """
        return self._get_optional_bool("Triggers Use Interpolated Strings")

    async def async_make_default(self):
        """Makes this profile the default profile."""
        await iterm2.rpc.async_set_default_profile(self.connection, self.guid)

class PartialProfile(Profile):
    """Represents a profile that has only a subset of fields available for reading."""

    @staticmethod
    async def async_get(
            connection: iterm2.connection.Connection,
            guids: typing.List[str]=None, properties: typing.List[str]=["Guid", "Name"]) -> 'PartialProfile':
        """Fetches a list of profiles.

        :param properties: Lists the properties to fetch. Pass None for all.
        :param guids: Lists GUIDs to list. Pass None for all profiles.

        :returns: A list of :class:`PartialProfile` objects with only the specified properties set.

        .. seealso::
            * Example ":ref:`theme_example`"
            * Example ":ref:`darknight`"
        """
        response = await iterm2.rpc.async_list_profiles(connection, guids, properties)
        profiles = []
        for responseProfile in response.list_profiles_response.profiles:
            profile = iterm2.profile.PartialProfile(None, connection, responseProfile.properties)
            profiles.append(profile)
        return profiles

    def __init__(self, session_id, connection, profile_property_list):
        """Initializes a PartialProfile from a profile_property_list protobuf."""
        super().__init__(session_id, connection, profile_property_list)

    async def async_get_full_profile(self) -> Profile:
        """Requests a full profile and returns it.

        Raises BadGUIDException if the Guid is not set or does not match a profile.

        :returns: A :class:`Profile`.

        .. seealso:: Example ":ref:`theme_example`"
        """
        if self.guid is None:
            raise
        response = await iterm2.rpc.async_list_profiles(self.connection, [self.guid], None)
        if len(response.list_profiles_response.profiles) != 1:
            raise BadGUIDException()
        return Profile(None, self.connection, response.list_profiles_response.profiles[0].properties)

    async def async_make_default(self):
        """Makes this profile the default profile."""
        await iterm2.rpc.async_set_default_profile(self.connection, self.guid)




"""Defines interfaces for registering functions."""
import asyncio
import inspect
import iterm2.notifications
import iterm2.rpc
import json
import traceback
import websockets

async def generic_handle_rpc(coro, connection, notif):
    rpc_notif = notif.server_originated_rpc_notification
    params = {}
    ok = False
    try:
        for arg in rpc_notif.rpc.arguments:
            name = arg.name
            if arg.HasField("json_value"):
                # NOTE: This can throw an exception if there are control characters or other nasties.
                value = json.loads(arg.json_value)
                params[name] = value
            else:
                params[name] = None
        result = await coro(**params)
        ok = True
    except KeyboardInterrupt as e:
        raise e
    except websockets.exceptions.ConnectionClosed as e:
        raise e
    except Exception as e:
        tb = traceback.format_exc()
        exception = { "reason": repr(e), "traceback": tb }
        await iterm2.rpc.async_send_rpc_result(connection, rpc_notif.request_id, True, exception)

    if ok:
        await iterm2.rpc.async_send_rpc_result(connection, rpc_notif.request_id, False, result)

class Reference:
    """Defines a reference to a variable for use in the @RPC decorator.

    .. seealso::
        * :func:`~iterm2.registration.RPC`.
        * Example ":ref:`mousemode_example`"
        * Example ":ref:`statusbar_example`"
    """
    def __init__(self, name):
        self.name = name

def RPC(func):
    """A decorator that adds an `async_register` value to the coroutine it decorates. `async_register` is a coroutine that will register the function as an RPC.

    An RPC is a function in a script that iTerm2 can invoke in response to some action, such a a keypress or a trigger. You use this decorator to register a coroutine as an RPC.

    Every RPC must have a unique signature. The signature is composed of two parts: first, the name, which comes from the name of the coroutine being decorated; second, the names of its arguments. The order of arguments is not important.

    The decorated coroutine will have a `async_register` value that you must call to complete the registration. `async_register` takes one required argument, the :class:`~iterm2.connection.Connection`. It also takes one optional argument, which is a timeout. The `timeout` is a value in seconds. If not given, the default timeout will be used. When waiting for an RPC to return, iTerm2 will stop waiting for the RPC after the timeout elapses.

    Do not use default values for arguments in your decorated coroutine, with one exception: a special kind of default value of type :class:`iterm2.Reference`. It names a variable that is visible in the context of the invocation. It will be transformed to the current value of that variable. This is the only way to get information about the current context. For example, a value of `~iterm2.Reference("id")` will give you the session ID of the context where the RPC was invoked. If the RPC is run from a keyboard shortcut, that is the ID of the session that had keyboard focus at the time of invocation.

    .. seealso::
        * Example ":ref:`badgetitle_example`"
        * Example ":ref:`blending_example`"
        * Example ":ref:`close_to_the_right_example`"
        * Example ":ref:`cls_example`"
        * Example ":ref:`jsonpretty_example`"
        * Example ":ref:`movetab_example`"

    That's complicated, but an example will make it clearer:

    Example:

      .. code-block:: python

          app = await iterm2.async_get_app(connection)

          @iterm2.RPC
          async def split_current_session_n_times(session_id=iterm2.Reference("id"), n=1):
              session = app.get_session_by_id(session_id)
              for i in range(n):
                  await session.async_split_pane()

          # Remember to call async_register!
          await split_current_session_n_times.async_register(connection)
    """
    async def async_register(connection, timeout=None):
        signature = inspect.signature(func)
        defaults = {}
        for k, v in signature.parameters.items():
            if isinstance(v.default, Reference):
                defaults[k] = v.default.name
        async def handle_rpc(connection, notif):
            await generic_handle_rpc(func, connection, notif)
        func.rpc_token = await iterm2.notifications.async_subscribe_to_server_originated_rpc_notification(
                connection,
                handle_rpc,
                func.__name__,
                signature.parameters.keys(),
                timeout,
                defaults,
                iterm2.notifications.RPC_ROLE_GENERIC)
        func.rpc_connection = connection

    func.async_register = async_register
    return func

def TitleProviderRPC(func):
    """A decorator (like :func:`~iterm2.registration.RPC`) that registers a session title provider.

    A session title provider is a function that gets called to compute the title of a session. It may be called frequently, whenever the session title is deemed to need recomputation. Once registered, it appears as an option in the list of title settings in preferences.

    It is called when any of its inputs change. This will only happen if one or more of the inputs are :func:`~iterm2.registration.Reference` references to variables in the session context.

    It must return a string.

    Note that the `async_register` function is different than in the :func:`~iterm2.registration.RPC` decorator: it takes three arguments. The first is the :class:`~iterm2.connection.Connection`. The second is a "display name", which is the string to show in preferences that the user may select to use this title provider. Then it takes a unique identifier, a string, which must be unique among all title providers.

    .. seealso:: Example ":ref:`georges_title_example`"

    Example:

      .. code-block:: python

          @iterm2.TitleProviderRPC
          async def upper_case_title(auto_name=iterm2.Reference("autoName?")):
              if not auto_name:
                  return ""
              return auto_name.upper()

          # Remember to call async_register!
          await upper_case_title.async_register(
                  connection,
                  display_name="Upper-case Title",
                  unique_identifier="com.iterm2.example.title-provider")
    """
    async def async_register(connection, display_name, unique_identifier, timeout=None):
        assert unique_identifier
        signature = inspect.signature(func)
        defaults = {}
        for k, v in signature.parameters.items():
            if isinstance(v.default, Reference):
                defaults[k] = v.default.name
        async def handle_rpc(connection, notif):
            await generic_handle_rpc(func, connection, notif)
        func.rpc_token = await iterm2.notifications.async_subscribe_to_server_originated_rpc_notification(
                connection=connection,
                callback=handle_rpc,
                name=func.__name__,
                arguments=signature.parameters.keys(),
                timeout_seconds=timeout,
                defaults=defaults,
                role=iterm2.notifications.RPC_ROLE_SESSION_TITLE,
                session_title_display_name=display_name,
                session_title_unique_id=unique_identifier)
        func.rpc_connection = connection

    func.async_register = async_register
    return func

def StatusBarRPC(func):
    """A decorator (like :func:`~iterm2.registration.RPC`) that registers a custom status bar component.

    See :class:`~iterm2.statusbar.StatusBarComponent` for details on what a status bar component is.

    The coroutine is called when any of its inputs change. This will only happen if one or more of the inputs are :func:`~iterm2.registration.Reference` references to variables in the session context.

    The coroutine *must* take an argument named `knobs` that will contain a dictionary with configuration settings.

    It may return a string or an array of strings. In the case that it returns an array, the longest string fitting the available space will be used.

    Note that unlike the other RPC decorators, you use :meth:`~iterm2.statusbar.StatusBarComponent.async_register` to register it, rather than a register property added to the coroutine.

    .. seealso::
        * Example ":ref:`escindicator_example`"
        * Example ":ref:`jsonpretty_example`"
        * Example ":ref:`mousemode_example`"
        * Example ":ref:`statusbar_example`"

    Example:

      .. code-block:: python

          component = iterm2.StatusBarComponent(
              short_description="Session ID",
              detailed_description="Show the session's identifier",
              knobs=[],
              exemplar="[session ID]",
              update_cadence=None,
              identifier="com.iterm2.example.statusbar-rpc")

          @iterm2.StatusBarRPC
          async def session_id_status_bar_coro(
                  knobs,
                  session_id=iterm2.Reference("id")):
              # This status bar component shows the current session ID, which
              # is useful for debugging scripts.
              return session_id

          @iterm2.RPC
          async def my_status_bar_click_handler(session_id):
              # When you click the status bar it opens a popover with the
              # message "Hello World"
              await component.async_open_popover(
                      session_id,
                      "Hello world",
                      iterm2.Size(200, 200))

          await component.async_register(
                  connection,
                  session_id_status_bar_coro,
                  onclick=my_status_bar_click_handler)
    """
    async def async_register(connection, component, timeout=None):
        signature = inspect.signature(func)
        defaults = {}
        for k, v in signature.parameters.items():
            if isinstance(v.default, Reference):
                defaults[k] = v.default.name

        async def wrapper(**kwargs):
            """handle_rpc->generic_handle_rpc->wrapper->func"""
            # Fix up knobs to not be JSON
            if "knobs" in kwargs:
                knobs_json = kwargs["knobs"]
                kwargs["knobs"] = json.loads(knobs_json)
            return await func(**kwargs)

        async def handle_rpc(connection, notif):
            """This gets run first."""
            await generic_handle_rpc(wrapper, connection, notif)

        func.rpc_token = await iterm2.notifications.async_subscribe_to_server_originated_rpc_notification(
                connection=connection,
                callback=handle_rpc,
                name=func.__name__,
                arguments=signature.parameters.keys(),
                timeout_seconds=timeout,
                defaults=defaults,
                role=iterm2.notifications.RPC_ROLE_STATUS_BAR_COMPONENT,
                status_bar_component=component)
        func.rpc_connection = connection

    func.async_register = async_register

    return func




"""Subscribe/unsubscribe from async notifications.

This module provides functions that let you subscribe and unsubscribe from notifications. iTerm2
posts notifications when some event of interest (for example, a keystroke) occurs. By subscribing to
a notifications your async callback will be run when the event occurs.
"""
import asyncio
import enum
import iterm2.api_pb2
import iterm2.connection
import iterm2.rpc
import iterm2.variables

RPC_ROLE_GENERIC = iterm2.api_pb2.RPCRegistrationRequest.Role.Value("GENERIC")
RPC_ROLE_SESSION_TITLE = iterm2.api_pb2.RPCRegistrationRequest.Role.Value("SESSION_TITLE")
RPC_ROLE_STATUS_BAR_COMPONENT = iterm2.api_pb2.RPCRegistrationRequest.Role.Value("STATUS_BAR_COMPONENT")

def _get_handlers():
    """Returns the registered notification handlers.

    :returns: (session, notification_type) -> [coroutine, ...]
    """
    if not hasattr(_get_handlers, 'handlers'):
        _get_handlers.handlers = {}
    return _get_handlers.handlers

## APIs -----------------------------------------------------------------------

class SubscriptionException(Exception):
    """Raised when a subscription attempt fails."""
    pass

async def async_unsubscribe(connection, token):
    """
    Unsubscribes from a notification.

    :param connection: A connected :class:`Connection`.
    :param token: The result of a previous subscribe call.
    """
    key, coro = token
    coros = _get_handlers()[key]
    coros.remove(coro)
    if coros:
        _get_handlers()[key] = coros
    else:
        del _get_handlers()[key]
        if len(key) == 2:
            session, notification_type = key
            await _async_subscribe(
                connection,
                False,
                notification_type,
                coro,
                session=session)
        else:
            if key[-1] == iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE:
                scope, identifier, name, notification_type = key
                request = iterm2.api_pb2.VariableMonitorRequest()
                request.name = name
                request.scope = scope
                request.identifier = identifier
                await _async_subscribe(
                        connection,
                        False,
                        notification_type,
                        coro,
                        key=key,
                        variable_monitor_request=request)
            else:
                # The key is custom and there should be code to handle unsubscribing as an elif clause here.
                assert False

async def async_subscribe_to_new_session_notification(connection, callback):
    """
    Registers a callback to be run when a new session is created.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an
      :class:`Connection` and iterm2.api_pb2.NewSessionNotification.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(connection, True, iterm2.api_pb2.NOTIFY_ON_NEW_SESSION, callback)

async def async_subscribe_to_keystroke_notification(connection, callback, session=None):
    """
    Registers a callback to be run when a key is pressed.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection` and
      iterm2.api_pb2.KeystrokeNotification.
    :param session: The session to monitor, or None.

    Returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_KEYSTROKE,
        callback,
        session=session,
        keystroke_monitor_request=None)

async def async_filter_keystrokes(connection, patterns_to_ignore, session=None):
    """
    Subscribe to a pseudo-notification that causes keystrokes matching patterns to be ignored.

    :param connection: A connected :class:`Connection`.
    :param patterns_to_ignore: A list of keystroke patterns that iTerm2 should not handle, expecting the script will handle it alone. Objects are class :class:`KeystrokePattern`.
    :param session: The session to monitor, or None.

    Returns: A token that can be passed to unsubscribe.
    """
    kmr = iterm2.api_pb2.KeystrokeMonitorRequest()
    kmr.patterns_to_ignore.extend(list(map(lambda x: x.to_proto(), patterns_to_ignore)))
    async def callback(connection, notif):
      pass
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.KEYSTROKE_FILTER,
        callback,
        session=session,
        keystroke_monitor_request=kmr)

async def async_subscribe_to_screen_update_notification(connection, callback, session=None):
    """
    Registers a callback to be run when the screen contents change.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection` and
      iterm2.api_pb2.ScreenUpdateNotification..
    :param session: The session to monitor, or None.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_SCREEN_UPDATE,
        callback,
        session=session)

async def async_subscribe_to_prompt_notification(connection, callback, session=None):
    """
    Registers a callback to be run when a shell prompt is received.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection` and
      iterm2.api_pb2.PromptNotification.
    :param session: The session to monitor, or None.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_PROMPT,
        callback,
        session=session)

async def async_subscribe_to_custom_escape_sequence_notification(connection,
                                                                 callback,
                                                                 session=None):
    """
    Registers a callback to be run when a custom escape sequence is received.

    The escape sequence is OSC 1337 ; Custom=id=<identity>:<payload> ST

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection` and
      iterm2.api_pb2.CustomEscapeSequenceNotification.
    :param session: The session to monitor, or None.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_CUSTOM_ESCAPE_SEQUENCE,
        callback,
        session=session)

async def async_subscribe_to_terminate_session_notification(connection, callback):
    """
    Registers a callback to be run when a session terminates.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection` and
      iterm2.api_pb2.TerminateSessionNotification.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_TERMINATE_SESSION,
        callback,
        session=None)

async def async_subscribe_to_layout_change_notification(connection, callback):
    """
    Registers a callback to be run when the relationship between sessions, tabs,
    and windows changes.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection` and
      iterm2.api_pb2.LayoutChangedNotification.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_LAYOUT_CHANGE,
        callback,
        session=None)

async def async_subscribe_to_focus_change_notification(connection, callback):
    """
    Registers a callback to be run when focus changes.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection` and
      iterm2.api_pb2.FocusChangedNotification.

    :returns: A token that can be passed to unsubscribe.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_FOCUS_CHANGE,
        callback,
        session=None)

async def async_subscribe_to_broadcast_domains_change_notification(connection, callback):
    """
    Registers a callback to be run when the current broadcast domains change.

    See also: :meth:`~iTerm2.App.parse_broadcast_domains`. Pass it `notification.broadcast_domains`.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection` and
      iterm2.api_pb2.BroadcastDomainsChangedNotification.
    """
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_BROADCAST_CHANGE,
        callback,
        session=None)

async def async_subscribe_to_server_originated_rpc_notification(
        connection,
        callback,
        name,
        arguments=[],
        timeout_seconds=5,
        defaults={},
        role=RPC_ROLE_GENERIC,
        session_title_display_name=None,
        session_title_unique_id=None,
        status_bar_component=None):
    """
    Registers a callback to be run when the server wants to invoke an RPC.

    You probably want to use :meth:`~iterm2.Registration.async_register_rpc_handler`
    instead of this. It's a much higher level API.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection` and iterm2.api_pb2.ServerOriginatedRPCNotification.
    :param timeout_seconds: How long iTerm2 should wait for this function to return or `None` to use the default timeout.
    :param defaults: Gives default values. Names correspond to argument names in `arguments`. Values are in-scope variables at the callsite.
    :param role: Defines the special purpose of this RPC. If none, use `RPC_ROLE_GENERIC`.
    :param session_title_display_name: Used by the `RPC_ROLE_SESSION_TITLE` role to give the name of the function to show in preferences.

    :returns: A token that can be passed to unsubscribe.
    """
    rpc_registration_request = iterm2.api_pb2.RPCRegistrationRequest()
    rpc_registration_request.name = name
    if timeout_seconds is not None:
        rpc_registration_request.timeout = timeout_seconds
    args = []
    for arg_name in arguments:
        arg = iterm2.api_pb2.RPCRegistrationRequest.RPCArgumentSignature()
        arg.name = arg_name
        args.append(arg)
    rpc_registration_request.arguments.extend(args)
    rpc_registration_request.role = role

    if len(defaults) > 0:
        d = []
        for name in defaults:
            assert name in arguments, "Name for default '{}' not in arguments {}".format(name, arguments)
            path = defaults[name]
            argd = iterm2.api_pb2.RPCRegistrationRequest.RPCArgument()
            argd.name = name
            argd.path = path
            d.append(argd)

        rpc_registration_request.defaults.extend(d)

    if session_title_display_name is not None:
        rpc_registration_request.session_title_attributes.display_name = session_title_display_name
        assert session_title_unique_id
        rpc_registration_request.session_title_attributes.unique_identifier = session_title_unique_id
    elif status_bar_component is not None:
        status_bar_component.set_fields_in_proto(rpc_registration_request.status_bar_component_attributes)

    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_SERVER_ORIGINATED_RPC,
        callback,
        rpc_registration_request=rpc_registration_request)

async def async_subscribe_to_variable_change_notification(connection, callback, scope, name, identifier):
    """
    Registers a callback to be invoked when a variable changes.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: an :class:`Connection` and iterm2.api_pb2.VariableChangedNotification.
    :param scope: A :class:`VariableScopes` enumerated value.
    :param name: The name of the variable, a string.
    :param identifier: The identifier of the object (window, tab, or session) being monitored, or None for app. Sometimes this will be "all" or "active".
    """
    request = iterm2.api_pb2.VariableMonitorRequest()
    request.name = name
    request.scope = scope
    if identifier is not None:
        request.identifier = identifier
    else:
        request.identifier = ""
    key = (request.scope, request.identifier, request.name, iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE)
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE,
        callback,
        session=None,
        variable_monitor_request=request,
        key=key)

async def async_subscribe_to_profile_change_notification(connection, callback, guid):
    """
    Register a callback to be invoked when a profile changes.

    :param connection: A connected :class:`Connection`.
    :param callback: A coroutine taking two arguments: a :class:`Connection` and iterm2.api_pb2.ProfileChangedNotification.
    :parm guid: The guid to monitor. A string.
    """
    request = iterm2.api_pb2.ProfileChangeRequest()
    request.guid = guid
    key = (guid, iterm2.api_PB2.NOTIFY_ON_PROFILE_CHANGE)
    return await _async_subscribe(
        connection,
        True,
        iterm2.api_PB2.NOTIFY_ON_PROFILE_CHANGE,
        callback,
        session=None,
        profile_change_request=request,
        key=key)



## Private --------------------------------------------------------------------

def _string_rpc_registration_request(rpc):
    """Converts ServerOriginatedRPC or RPCSignature to a string."""
    if rpc is None:
        return None
    args = sorted(map(lambda x: x.name, rpc.arguments))
    return rpc.name + "(" + ",".join(args) + ")"

async def _async_subscribe(connection, subscribe, notification_type, callback, session=None, rpc_registration_request=None, keystroke_monitor_request=None, variable_monitor_request=None, key=None, profile_change_request=None):
    _register_helper_if_needed()
    transformed_session = session if session is not None else "all"

    # Register locally in case iTerm2 responds with an RPC immediately.
    # That is typical when registering a session title provider when a session is already open.
    if subscribe:
        if key:
            _register_notification_handler_impl(key, callback)
        else:
            _register_notification_handler(
                    session,
                    _string_rpc_registration_request(rpc_registration_request),
                    notification_type,
                    callback)

    response = await iterm2.rpc.async_notification_request(
        connection,
        subscribe,
        notification_type,
        transformed_session,
        rpc_registration_request,
        keystroke_monitor_request,
        variable_monitor_request,
        profile_change_request)
    status = response.notification_response.status
    status_ok = (status == iterm2.api_pb2.NotificationResponse.Status.Value("OK"))

    unregister = False
    if subscribe:
        already = (status == iterm2.api_pb2.NotificationResponse.Status.Value("ALREADY_SUBSCRIBED"))
        if status_ok or already:
            if key:
                return (key, callback)
            else:
                return ((session, notification_type), callback)
        else:
            # Uh oh, undo the registration.
            unregister = True
    else:
        # Unsubscribe
        if status_ok:
            unregister = True

    if unregister:
        if key:
            _unregister_notification_handler_impl(key, callback)
        else:
            _unregister_notification_handler(
                session,
                _string_rpc_registration_request(rpc_registration_request),
                notification_type,
                callback)

        if not subscribe and status_ok:
            # Normal code path for unsubscribe
            return

    raise SubscriptionException(iterm2.api_pb2.NotificationResponse.Status.Name(status))

def _register_helper_if_needed():
    if not hasattr(_register_helper_if_needed, 'haveRegisteredHelper'):
        _register_helper_if_needed.haveRegisteredHelper = True
        iterm2.connection.Connection.register_helper(_async_dispatch_helper)

async def _async_dispatch_helper(connection, message):
    handlers, sub_notification = _get_notification_handlers(message)
    for handler in handlers:
        assert handler is not None
        await handler(connection, sub_notification)

    if not handlers and message.notification.HasField('server_originated_rpc_notification'):
        # If we get an RPC we haven't registered for handle the error because
        # otherwise it has to time out. If you get here there is probably a bug.
        key, _sub_notification = _get_handler_key_from_notification(message.notification)
        exception = { "reason": "No such function: {} (key was {})".format(_string_rpc_registration_request(message.notification.server_originated_rpc_notification.rpc), key) }
        await iterm2.rpc.async_send_rpc_result(connection, message.notification.server_originated_rpc_notification.request_id, True, exception)

    return bool(handlers)

def _get_all_sessions_handler_key_from_notification(notification):
    if notification.HasField('variable_changed_notification'):
        return (iterm2.api_pb2.VariableScope.Value("SESSION"),
                "all",
                notification.variable_changed_notification.name,
                iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE)
    else:
        standard_key, _ = _get_handler_key_from_notification(notification)
        return (None, standard_key[1])

    return None


def _get_handler_key_from_notification(notification):
    key = None

    if notification.HasField('keystroke_notification'):
        key = (notification.keystroke_notification.session, iterm2.api_pb2.NOTIFY_ON_KEYSTROKE)
        notification = notification.keystroke_notification
    elif notification.HasField('screen_update_notification'):
        key = (notification.screen_update_notification.session,
               iterm2.api_pb2.NOTIFY_ON_SCREEN_UPDATE)
        notification = notification.screen_update_notification
    elif notification.HasField('prompt_notification'):
        key = (notification.prompt_notification.session, iterm2.api_pb2.NOTIFY_ON_PROMPT)
        notification = notification.prompt_notification
    elif notification.HasField('location_change_notification'):
        key = (notification.location_change_notification.session,
               iterm2.api_pb2.NOTIFY_ON_LOCATION_CHANGE)
        notification = notification.location_change_notification
    elif notification.HasField('custom_escape_sequence_notification'):
        key = (notification.custom_escape_sequence_notification.session,
               iterm2.api_pb2.NOTIFY_ON_CUSTOM_ESCAPE_SEQUENCE)
        notification = notification.custom_escape_sequence_notification
    elif notification.HasField('new_session_notification'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_NEW_SESSION)
        notification = notification.new_session_notification
    elif notification.HasField('terminate_session_notification'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_TERMINATE_SESSION)
        notification = notification.terminate_session_notification
    elif notification.HasField('layout_changed_notification'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_LAYOUT_CHANGE)
        notification = notification.layout_changed_notification
    elif notification.HasField('focus_changed_notification'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_FOCUS_CHANGE)
        notification = notification.focus_changed_notification
    elif notification.HasField('server_originated_rpc_notification'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_SERVER_ORIGINATED_RPC, _string_rpc_registration_request(notification.server_originated_rpc_notification.rpc))
    elif notification.HasField('broadcast_domains_changed'):
        key = (None, iterm2.api_pb2.NOTIFY_ON_BROADCAST_CHANGE)
    elif notification.HasField('variable_changed_notification'):
        key = (notification.variable_changed_notification.scope,
               notification.variable_changed_notification.identifier,
               notification.variable_changed_notification.name,
               iterm2.api_pb2.NOTIFY_ON_VARIABLE_CHANGE)
        notification = notification.variable_changed_notification
    elif notification.HasField('profile_changed_notification'):
        key = (notification.profile_changed_notification.guid,
                iterm2.api_pb2.NOTIFY_ON_PROFILE_CHANGE)
    return key, notification

def _get_notification_handlers(message):
    key, sub_notification = _get_handler_key_from_notification(message.notification)
    if key is None:
        return ([], None)

    # This fallback catches "all"-style subscriptions.
    fallback = _get_all_sessions_handler_key_from_notification(message.notification)

    if key in _get_handlers():
        return (_get_handlers()[key], sub_notification)
    elif fallback in _get_handlers():
        return (_get_handlers()[fallback], sub_notification)
    return ([], None)

def _register_notification_handler(session, rpc_registration_request, notification_type, coro):
    assert coro is not None

    if rpc_registration_request is None:
        key = (session, notification_type)
    else:
        key = (session, notification_type, rpc_registration_request)
    _register_notification_handler_impl(key, coro)

def _register_notification_handler_impl(key, coro):
    if key in _get_handlers():
        _get_handlers()[key].append(coro)
    else:
        _get_handlers()[key] = [coro]

def _unregister_notification_handler(session, rpc_registration_request, notification_type, coro):
    assert coro is not None

    if rpc_registration_request is None:
        key = (session, notification_type)
    else:
        key = (session, notification_type, rpc_registration_request)
    _unregister_notification_handler_impl(key, coro)

def _unregister_notification_handler_impl(key, coro):
    if key in _get_handlers():
        if coro in _get_handlers()[key]:
            _get_handlers()[key].remove(coro)




"""Provides interfaces for interacting with selected text regions."""
import enum
import iterm2.api_pb2
import iterm2.connection
import iterm2.screen
import iterm2.util
import itertools
from operator import itemgetter
import typing

class SelectionMode(enum.Enum):
    """Enumerated list of modes for selecting text."""
    CHARACTER = 0  #: character-by-character selection
    WORD = 1  #: word-by-word selection
    LINE = 2  #: row-by-row selection
    SMART = 3  #: smart selection
    BOX = 4  #: rectangular region
    WHOLE_LINE = 5  #: entire wrapped lines, which could occupy many rows

    @staticmethod
    def fromProtoValue(p):
        global MODE_MAP
        return MODE_MAP[p]

    @staticmethod
    def toProtoValue(p):
        global INVERSE_MODE_MAP
        return INVERSE_MODE_MAP[p]

class SubSelection:
    """Describes a continguous block of selected characters.

    :param windowedCoordRange: A :class:`~iterm2.util.WindowedCoordRange` describing the range.
    :param mode: A :class:`SelectionMode` describing how the selection is interpreted and extended.
    :param connected: If true, no newline exists between this and the next sub-selection.
    """
    def __init__(
            self,
            windowedCoordRange: iterm2.util.WindowedCoordRange,
            mode:SelectionMode,
            connected: bool):
        self.__windowedCoordRange = windowedCoordRange
        self.__mode = mode
        self.__connected = connected

    @property
    def windowedCoordRange(self) -> iterm2.util.WindowedCoordRange:
        return self.__windowedCoordRange

    @property
    def mode(self) -> SelectionMode:
        return self.__mode

    @property
    def proto(self):
        p = iterm2.api_pb2.SubSelection()
        p.windowed_coord_range.CopyFrom(self.__windowedCoordRange.proto)
        p.selection_mode = SelectionMode.toProtoValue(self.__mode)
        return p

    @property
    def connected(self) -> bool:
        return self.__connected

    async def async_get_string(self, connection: iterm2.connection.Connection, session_id: str) -> str:
        """Gets the text belonging to this subselection.

        :param connection: The connection to iTerm2.
        :param session_id: The ID of the session for which to look up the selected text.
        """
        result = await iterm2.rpc.async_get_screen_contents(
            connection,
            session_id,
            self.__windowedCoordRange)
        if result.get_buffer_response.status == iterm2.api_pb2.GetBufferResponse.Status.Value("OK"):
            screenContents = iterm2.screen.ScreenContents(result.get_buffer_response)
            s = ""
            wcr = screenContents.windowed_coord_range
            i = 0
            while i < screenContents.number_of_lines:
                line = screenContents.line(i)
                i += 1
                s += line.string
                if line.hard_eol:
                    s += "\n"
            return s
        else:
            raise iterm2.rpc.RPCException(iterm2.api_pb2.GetBufferResponse.Status.Name(result.get_buffer_response.status))

    def enumerateRanges(self, callback):
        if self.__windowedCoordRange.hasWindow:
            right = self.__windowedCoordRange.right
            startX = self.__windowedCoordRange.start.x
            y = self.__windowedCoordRange.coordRange.start.y
            while y < self.__windowedCoordRange.coordRange.end.y:
                callback(iterm2.util.CoordRange(
                    iterm2.util.Point(startX, y),
                    iterm2.util.Point(right, y)))
                startX = self.__windowedCoordRange.left

                y += 1

            callback(iterm2.util.CoordRange(
                iterm2.util.Point(
                    startX,
                    self.__windowedCoordRange.coordRange.end.y),
                iterm2.util.Point(
                    self.__windowedCoordRange.end.x,
                    self.__windowedCoordRange.coordRange.end.y)))
        else:
            callback(self.__windowedCoordRange.coordRange)

class Selection:
    """A collection of :class:`SubSelection` objects, describing all the selections in a session.

    :param subSelections: An array of :class:`SubSelection` objects.
    """
    def __init__(self, subSelections: typing.List[SubSelection]):
        self.__subSelections = subSelections

    @property
    def subSelections(self) -> typing.List[SubSelection]:
        return self.__subSelections

    async def _async_get_content_in_range(self, connection, session_id, coordRange):
        result = await iterm2.rpc.async_get_screen_contents(
            connection,
            session_id,
            coordRange)
        if result.get_buffer_response.status == iterm2.api_pb2.GetBufferResponse.Status.Value("OK"):
            screenContents = iterm2.screen.ScreenContents(result.get_buffer_response)
            s = ""
            wcr = screenContents.windowed_coord_range
            i = 0
            while i < screenContents.number_of_lines:
                line = screenContents.line(i)
                i += 1
                s += line.string
                if line.hard_eol:
                    s += "\n"
            return s

    async def async_get_string(self, connection: iterm2.connection.Connection, session_id: str, width: int) -> str:
        """Returns the selected text.

        :param connection: The connection to iTerm2.
        :param session_id: The ID of the session for which to look up the selected text.
        :param width: The width (number of columns) of the session.
        """
        if len(self.__subSelections) == 1:
            return await self.__subSelections[0].async_get_string(connection, session_id)

        result = ""
        async def handleRange(coordRange, eol):
            content = await self._async_get_content_in_range(connection, session_id, coordRange)
            nonlocal result
            result += content
            if eol and not content.endswith("\n"):
                result += "\n"
        await self.async_enumerate_ranges(connection, session_id, width, handleRange)
        return result

    async def async_enumerate_ranges(self, connection, session_id, width, cb):
        """Gets the text belonging to each subselection and concatenates them with newlines.

        :param connection: A :class:`~iterm2.connection.Connection`.
        :param session_id: A string session ID.

        :returns: A string with the selected text.
        """
        if len(self.__subSelections) == 0:
            return

        # Ranges ending at connectors don't get a newline following.
        connectors = set()
        indexes = set()
        for outer in self.__subSelections:
            if outer.connected:
                thePosition = outer.windowedCoordRange.coordRange.end.x + outer.windowedCoordRange.coordRange.end.y * width;
                connectors |= {thePosition}

            theRange = iterm2.util.Range(0, 0)
            def handleRange(outerRange):
                nonlocal indexes
                nonlocal theRange
                theRange = iterm2.util.Range(
                        outerRange.start.x + outerRange.start.y * width,
                        outerRange.length(width));

                indexesToAdd = theRange.toSet
                indexesToRemove = set()
                def f(values):
                    i,x=values
                    return i-x
                # Iterate contiguous ranges
                indexes_of_interest = indexes.intersection(indexesToAdd)
                for k, g in itertools.groupby(enumerate(sorted(indexes_of_interest)), f):
                    # range exists in both indexes and theRange
                    values_in_range=map(itemgetter(1), g)
                    indexesToRemove |= set(values_in_range)
                    indexesToAdd -= set(values_in_range)
                indexes -= indexesToRemove
                indexes |= indexesToAdd

                # In multipart windowed ranges, add connectors for the endpoint of all but the last
                # range. Each enumerated range is on its own line.
                if (outer.windowedCoordRange.hasWindow and
                    outerRange.end == outer.windowedCoordRange.coordRange.end and
                    theRange.length > 0):
                    connectors |= {theRange.max}

            outer.enumerateRanges(handleRange)

        # the ranges may be out of order so put them in an array and then sort it.
        allRanges = []
        def f(values):
            i,x=values
            return i-x
        # Iterate contiguous ranges
        for k, g in itertools.groupby(enumerate(sorted(indexes)), f):
            # range exists in both indexes and theRange
            values_in_range=list(map(itemgetter(1), g))
            coordRange = iterm2.util.CoordRange(
                    iterm2.util.Point(
                        min(values_in_range) % width,
                        min(values_in_range) // width),
                    iterm2.util.Point(
                        max(values_in_range) % width,
                        max(values_in_range) // width))
            allRanges.append(coordRange)

        def rangeKey(r):
            return r.start.y * width + r.start.x
        sortedRanges = sorted(allRanges, key=rangeKey)
        for idx, theRange in enumerate(sortedRanges):
            endIndex = theRange.start.x + theRange.start.y * width + theRange.length(width)
            eol = (endIndex not in connectors) and idx + 1 < len(sortedRanges)
            stop = await cb(iterm2.util.WindowedCoordRange(theRange), eol)
            if stop:
                break


MODE_MAP = {
        iterm2.api_pb2.SelectionMode.Value("CHARACTER"): SelectionMode.CHARACTER,
        iterm2.api_pb2.SelectionMode.Value("WORD"): SelectionMode.WORD,
        iterm2.api_pb2.SelectionMode.Value("LINE"): SelectionMode.LINE,
        iterm2.api_pb2.SelectionMode.Value("SMART"): SelectionMode.SMART,
        iterm2.api_pb2.SelectionMode.Value("BOX"): SelectionMode.BOX,
        iterm2.api_pb2.SelectionMode.Value("WHOLE_LINE"): SelectionMode.WHOLE_LINE }

INVERSE_MODE_MAP = {v: k for k, v in MODE_MAP.items()}




import enum

class ColorSpace(enum.Enum):
    """Describes the color space of a :ref:`Color`."""
    SRGB="sRGB" #: SRGB color space
    CALIBRATED="Calibrated"  #: Device color space

class Color:
    """Describes a color.

      :param r: Red, in 0-255
      :param g: Green, in 0-255
      :param b: Blue, in 0-255
      :param a: Alpha, in 0-255
      :param color_space: The color space. Only sRGB is supported currently.
      """
    def __init__(self, r: int=0, g: int=0, b: int=0, a: int=255, color_space: ColorSpace=ColorSpace.SRGB):
        """Create a color."""
        self.__red = r
        self.__green = g
        self.__blue = b
        self.__alpha = a
        self.__color_space = color_space

    def __repr__(self):
        return "({},{},{},{} {})".format(
            round(self.red),
            round(self.green),
            round(self.blue),
            round(self.alpha),
            self.color_space)

    @property
    def red(self) -> int:
        """The color's red component."""
        return self.__red

    @red.setter
    def red(self, value: float):
        """Sets the color's red component."""
        self.__red = value

    @property
    def green(self) -> float:
        """The color's green component."""
        return self.__green

    @green.setter
    def green(self, value: float):
        """Sets the color's green component."""
        self.__green = value

    @property
    def blue(self) -> float:
        """The color's blue component."""
        return self.__blue

    @blue.setter
    def blue(self, value: float):
        """Sets the color's blue component."""
        self.__blue = value

    @property
    def alpha(self) -> float:
        """The color's alpha component."""
        return self.__alpha

    @alpha.setter
    def alpha(self, value: float):
        """Sets the color's alpha component."""
        self.__alpha = value

    @property
    def color_space(self) -> ColorSpace:
        """The color's color space."""
        return self.__color_space

    @color_space.setter
    def color_space(self, value: ColorSpace):
        """Sets the color's scolor space."""
        self.__color_space = value

    def get_dict(self):
        """Returns a dictionary representation of this color.

        Suitable for conversion to a JSON object to pass to iTerm2."""
        return {
            "Red Component": self.red / 255.0,
            "Green Component": self.green / 255.0,
            "Blue Component": self.blue / 255.0,
            "Alpha Component": self.alpha / 255.0,
            "Color Space": self.color_space.value
            }

    def from_dict(self, input_dict):
        """Updates the color from the dictionary's contents."""
        self.red = float(input_dict["Red Component"]) * 255
        self.green = float(input_dict["Green Component"]) * 255
        self.blue = float(input_dict["Blue Component"]) * 255
        if "Alpha Component" in input_dict:
            self.alpha = float(input_dict["Alpha Component"]) * 255
        else:
            self.alpha = 255
        if "Color Space" in input_dict:
            self.color_space = ColorSpace(input_dict["Color Space"])
        else:
            # This is the default because it is what profiles use by default.
            self.color_space = ColorSpace.CALIBRATED

    @property
    def json(self):
        return json.dumps(self.get_dict())

